# Python Imports
import logging
import traceback
import datetime

# Third Party Imports
import numpy as np

# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtWidgets import QWidget
from PyQt5.QtCore import pyqtSignal

# Own Imports
from kmap import __directory__
from kmap.controller.sliceddatatab import SlicedDataTab
from kmap.controller.orbitaldatatab import OrbitalDataTab
from kmap.controller.profileplottab import ProfilePlotTab
from kmap.controller.renametabwindow import RenameTabWindow
from kmap.controller.lmfitplottab import LMFitPlotTab
from kmap.controller.splitviewtab import SplitViewTab
from kmap.controller.lmfittab import LMFitTab, LMFitResultTab
from kmap.controller.filetab import FileViewerTab, FileEditorTab
from kmap.library.qwidgetsub import Tab
from kmap.config.config import config

# Load .ui File
UI_file = __directory__ / 'ui/tabwidget.ui'
TabWidget_UI, _ = uic.loadUiType(UI_file)


class TabWidget(QWidget, TabWidget_UI):
    tab_added = pyqtSignal(Tab)

    def __init__(self, *args, **kwargs):
        # Setup GUI
        super(TabWidget, self).__init__(*args, **kwargs)
        self.setupUi(self)
        self._connect()

    def open_sliced_data_tab(self, path):
        # Opens a new sliced data tab

        log = logging.getLogger('kmap')
        log.info('Trying to load %s' % str(path))

        try:
            tab = SlicedDataTab.init_from_path(str(path))
            title = tab.get_title()
            tooltip = tab.to_string()
            self._open_tab(tab, title, tooltip)

        except Exception as e:
            log.error('Couldn\'t load %s' % str(path))
            log.error(traceback.format_exc())

    def open_sliced_data_tab_by_URLs(self, URLs):
        try:
            tab = SlicedDataTab.init_from_URLs(URLs)
            title = tab.get_title()
            tooltip = tab.to_string()
            self._open_tab(tab, title, tooltip)

        except ValueError as e:
            pass

        except Exception as e:
            log = logging.getLogger('kmap')
            log.error('Couldn\'t load URLs')
            log.error(traceback.format_exc())

    def open_sliced_data_tab_by_URL(self, URL):
        try:
            tab = SlicedDataTab.init_from_URL(URL)
            title = tab.get_title()
            tooltip = tab.to_string()
            self._open_tab(tab, title, tooltip)

        except ValueError as e:
            pass

        except Exception as e:
            log = logging.getLogger('kmap')
            log.error('Couldn\'t load URL')
            log.error(traceback.format_exc())

    def open_sliced_data_tab_by_cube(self, URL):
        try:
            tab = SlicedDataTab.init_from_cube(URL)
            title = tab.get_title()
            tooltip = tab.to_string()
            self._open_tab(tab, title, tooltip)

        except ValueError as e:
            pass

        except Exception as e:
            log = logging.getLogger('kmap')
            log.error('Couldn\'t load cube file')
            log.error(traceback.format_exc())

    def open_orbital_data_tab(self):
        # Opens a new orbital data tab

        tab = OrbitalDataTab()
        tab.get_energy.connect(self.get_sliced_energy)
        title = tab.get_title()

        self._open_tab(tab, title)

        return tab

    def get_sliced_energy(self):
        match_type = config.get_key('orbital', 'match_type')
        tabs = self.get_tabs_of_type(SlicedDataTab)
        if tabs:
            axis_idx = tabs[0].get_axis()
            axis = tabs[0].get_data().axes[axis_idx]
            if match_type == 'average':
                energy = np.nanmean(axis.axis)
            elif match_type == 'max':
                energy = max(axis.axis)
            elif match_type == 'min':
                energy = min(axis.axis)
            else:
                slice = tabs[0].get_slice()
                energy = axis.axis[slice]

            self.sender().change_kinetic_energy(energy)

    def open_file_tab(self, path, title, editable=False, richText=False):
        if editable:
            tab = FileEditorTab(path, title)
        else:
            tab = FileViewerTab(path, title, richText=richText)

        self._open_tab(tab, title)

    def open_split_view_tab(self, sliced_tab, orbital_tab, save=None):
        if save is None:
            tab = SplitViewTab(sliced_tab, orbital_tab)

        else:
            tab = SplitViewTab.init_from_save(save, sliced_tab, orbital_tab)

        tab.locked_tabs = [sliced_tab, orbital_tab]

        sliced_tab.lock_while_open(tab)
        orbital_tab.lock_while_open(tab)
        tab.sliced_data_tab = sliced_tab
        tab.orbital_data_tab = orbital_tab

        self._open_tab(tab, 'Split View')

        return tab

    def open_lmfit_tab(self, sliced_tab, orbital_tab, save=None):
        if save is None:
            tab = LMFitTab(sliced_tab, orbital_tab)
            tab.title = 'LM-Fit Tab'

        else:
            tab = LMFitTab.init_from_save(save, sliced_tab, orbital_tab)

        tab.fit_finished.connect(self.open_result_tab)
        tab.locked_tabs = [sliced_tab, orbital_tab]

        sliced_tab.lock_while_open(tab)
        orbital_tab.lock_while_open(tab)
        tab.sliced_data_tab = sliced_tab
        tab.orbital_data_tab = orbital_tab

        self._open_tab(tab, 'LM-Fit Tab')

        return tab

    def open_result_tab(self, *args, sender=None, save=None, ID_map=None):
        lmfit_tab = self.sender() if sender is None else sender

        if save is None:
            tab = LMFitResultTab(lmfit_tab, *args)
        else:
            tab = LMFitResultTab.init_from_save(save, lmfit_tab, ID_map)

        tab.open_plot_tab.connect(self.open_lmfit_plot_tab)
        tab.locked_tabs = [lmfit_tab]

        lmfit_tab.lock_while_open(tab)

        current_time = datetime.datetime.now()
        title = 'Results (%i:%i)' % (current_time.hour, current_time.minute)
        tab.set_title(title)
        self._open_tab(tab, title)

        return tab

    def open_lmfit_plot_tab(self, *args, sender=None, save=None):
        result_tab = self.sender() if sender is None else sender

        if save is None:
            tab = LMFitPlotTab(*args, result_tab)

        else:
            tab = LMFitPlotTab.init_from_save(save, result_tab)

        tab.locked_tabs = [result_tab]

        result_tab.lock_while_open(tab)

        title = 'Plot'
        tab.set_title(title)
        self._open_tab(tab, title)

        return tab

    def open_profile_tab(self):
        tab = ProfilePlotTab(self, 'Profile Plot')

        self._open_tab(tab, 'Profile Plot')

        return tab

    def get_index_of(self, tab):
        return self.tab_widget.indexOf(tab)

    def get_orbital_tab_to_load_to(self):
        tab = None
        aux = self.tab_widget.currentWidget()
        if type(aux) == OrbitalDataTab:
            tab = aux

        else:
            for index in range(self.tab_widget.count()):
                aux = self.tab_widget.widget(index)

                if type(aux) == OrbitalDataTab:
                    tab = aux
                    break

        if tab is None:
            tab = self.open_orbital_data_tab()

        return tab

    def get_current_tab(self):
        return self.tab_widget.currentWidget()

    def get_all_tabs(self):
        count = self.tab_widget.count()
        tabs = [self.tab_widget.widget(i) for i in range(count)]

        return tabs

    def get_tabs_of_type(self, type_):
        tabs = []

        for index in range(self.tab_widget.count()):
            tab = self.tab_widget.widget(index)

            if type(tab) == type_ or type_ is None:
                tabs.append(tab)

        return tabs

    def rename_current_tab(self):
        self.rename_tab = RenameTabWindow()
        self.rename_tab.title_chosen.connect(self.set_tab_title)

    def set_tab_title(self, title):
        current_tab = self.get_current_tab()
        current_tab_index = self.tab_widget.indexOf(current_tab)

        current_tab.set_title(title)
        self.tab_widget.setTabText(current_tab_index, title)

        for tab in self.get_tabs_of_type(ProfilePlotTab):
            tab.rename_loaded_tab(current_tab, title)

    def duplicate_tab(self):
        current_tab = self.get_current_tab()

        if current_tab is None:
            return

        save = current_tab.save_state()

        if isinstance(current_tab, ProfilePlotTab):
            tab = self.open_profile_tab()
            tab.restore_state(save)

        elif isinstance(current_tab, OrbitalDataTab):
            tab = self.open_orbital_data_tab()
            tab.restore_state(save[0])

        elif isinstance(current_tab, SlicedDataTab):
            tab, _ = SlicedDataTab.init_from_save(save[0])

        elif isinstance(current_tab, LMFitResultTab):
            tab = self.open_result_tab(sender=save[1][0], save=save[0])

        elif isinstance(current_tab, LMFitTab):
            tab = self.open_lmfit_tab(*save[1], save=save[0])

        elif isinstance(current_tab, SplitViewTab):
            tab = self.open_split_view_tab(*save[1], save=save[0])

        elif isinstance(current_tab, LMFitPlotTab):
            tab = self.open_lmfit_plot_tab(sender=save[1][0], save=save[0])

        else:
            tab = type(current_tab).init_from_save(save[0])

        title = tab.get_title()
        tab.set_title(title)
        self._open_tab(tab, title)

    def open_tab_by_save(self, tab_save, *args):
        tab_type, save = tab_save

        if tab_type == 'ProfilePlotTab':
            tab = self.open_profile_tab()
            tab.restore_state(save)

        elif tab_type == 'OrbitalDataTab':
            tab = self.open_orbital_data_tab()
            ID_map = tab.restore_state(save)

        elif tab_type == 'LMFitTab':
            tab = self.open_lmfit_tab(args[0], args[1], save=save)

        elif tab_type == 'SplitViewTab':
            tab = self.open_split_view_tab(args[0], args[1], save=save)

        elif tab_type == 'LMFitResultTab':
            tab = self.open_result_tab(
                sender=args[0], save=save, ID_map=args[1])

        elif tab_type == 'LMFitPlotTab':
            tab = self.open_lmfit_plot_tab(sender=args[0], save=save)

        elif tab_type == 'SlicedDataTab':
            tab, ID_map = SlicedDataTab.init_from_save(save, *args)

        else:
            try:
                tab = eval(tab_type).init_from_save(save, *args)

            except:
                raise ValueError

        title = save['title'] if 'title' in save.keys() else tab.get_title()
        tab.set_title(title)
        self._open_tab(tab, title)

        if isinstance(tab, SlicedDataTab) or isinstance(tab, OrbitalDataTab):
            return tab, ID_map

        else:
            return tab

    def close_tab(self, index):
        # Close tab specified with index

        widget = self.tab_widget.widget(index)

        if widget.lock_tab is None:
            for tab in widget.locked_tabs:
                tab.unlock()

            self.tab_widget.removeTab(index)
            widget.close()
            widget.deleteLater()

        else:
            log = logging.getLogger('kmap')
            log.warning(
                'Tab is locked open because different tab still references it.')

    def _open_tab(self, tab, title, tooltip=None, index=None):
        index = self.tab_widget.addTab(tab, title)

        self.tab_widget.setCurrentIndex(index)

        if tooltip is not None:
            self.tab_widget.setTabToolTip(index, tooltip)

        self.tab_added.emit(tab)

    def _connect(self):
        self.tab_widget.tabCloseRequested.connect(self.close_tab)
        self.tab_widget.tabBarDoubleClicked.connect(self.rename_current_tab)



"""Provides various useful methods for any part of kMap.py.

This file defines a multitude of methods not part of any specific part
of kMap.py. Those methods are intended for use throughout kMap.py to
reduce the amount of copy-paste code.
"""

# Python Imports
import re
from math import isclose

# Third Party Imports
import numpy as np


def round_to(x, base):
    """Rounds 'x' to the nearest multiple of 'base'.

    Args:
        x (float): Number to be rounded.
        base (float): Base to be rounded to.

    Returns:
        (float): Rounded number.

    """

    return base * round(x / base)


def idx_closest_value(axis, value, decimals=5, bounds_error=True):
    """Returns the index of the value in 'axis' closest to the 'value'.

    Args:
        axis (list): Axis of which the closest value is to be found.
            Needs to be a regular grid without NaN values.
        value (float): Value the closes axis point is to be found.
        decimals (int): Decimals places float values get rounded to
            to avoid machine error.
    Returns:
        (int): Index of the axis element closes to 'value'.

    """

    # Shift the axis to start at 0 -> rounding to step size (=base)
    shift = axis[0]
    base = axis[1] - axis[0]
    mapped_value = round_to(value - shift, base) + shift

    try:
        # Round to fifth decimal place because machine error
        idx = list(np.around(axis, decimals=decimals)).index(
            np.around(mapped_value, decimals=decimals))

    except ValueError:
        # Value not found in list -> larger (smaller) than max (min)
        # + (-) step_size/2

        if bounds_error:
            return None

        else:
            if value > max(axis):
                idx = list(axis).index(max(axis))

            else:
                idx = list(axis).index(min(axis))

    return idx


def centered_meshgrid(x_axis, x_shift, y_axis, y_shift):
    """Returns a meshgrid made of 'x_axis' and 'y_axis' shifted by
    (x_shift, y_shift).

    Args:
        x_axis (list): x-axis for the meshgrid. Needs to be a regular
            grid without NaN values.
        x_shift (float): Shift for the x-axis.
        y_axis (list): y-axis for the meshgrid. Needs to be a regular
            grid without NaN values.
        y_shift (float): Shift for the y-axis.

    Returns:
        (list, list.T): Shifted meshgrides for the x and y axes shifted.
    """

    X = np.copy(x_axis).reshape(1, len(x_axis)) - x_shift
    Y = np.copy(y_axis).reshape(len(y_axis), 1) - y_shift

    return X, Y


def distance_in_meshgrid(X, Y):
    """Returns the grid entered filled with the distance to its center.

    Args:
        X (list): x-axis for the meshgrid. Needs to be a regular
            grid without NaN values.
        Y (list.T): y-axis for the meshgrid. Needs to be a regular
            grid without NaN values.

    Returns:
        (np.array): 2D numpy array with same size as grid. Each element
            carries the distance of its position in the grid.
    """

    return np.sqrt(X**2 + Y**2)


def normalize(data):
    """Normalizes the data to the number of non NaN values in the data.

    Args:
        data (list): Arbitrary list of any shape with numbers to be
            normalized. Can contain NaN values.

    Returns:
        (np.array): Numpy array of the same size as input but each value
            normalized.
    """

    data = np.array(data)
    num_elements = len(data[~np.isnan(data)])

    if num_elements == 0:
        norm = np.nan

    else:
        norm = np.nansum(data) / num_elements

    return norm


def orientation_to_euler_angle(orientation):
    """Returns the Euler angles for the plane specified in 'orientation'.

    Args:
        orientation (string): Plane for which the Euler angles are to be
            return. Possible values: 'xy', 'yx', 'xz', 'zx'

    Returns:
        (list): List of length three containing the Euler angles in the
            following order: Phi, Theta, Psi
    """

    phi = 0 if orientation in ['xy', 'xz', 'zx'] else 90
    theta = 0 if orientation in ['xy', 'yx'] else 90
    psi = 0 if orientation in ['xy', 'yx', 'yz', 'xz'] else 90

    return phi, theta, psi


def axis_from_range(range_, num):
    """Calculates a full 1D axis from range values and number of
        elements. Result can be used for interpolation method.

    Args:
        range_ (float): 1D list with min and max value (inclusive).
        num (int): Number of points.

    Returns:
        (float): 1D np.array containing the resulting axis.

    """

    return np.linspace(range_[0], range_[1], num=num, endpoint=True,
                       dtype=np.float64)


def range_from_axes(*args):
    """Returns a range list for both axis.

    Args:
        *args (list): Arbitrary many axis. Each axis has to be a 1D list
            with no NaN values.

    Returns:
        (list): List of ranges for each axis supplied.
        (list): List of step sizes for each axis supplied.
    """

    range_ = []
    step_size = []

    for axis in args:
        range_.append([axis[0], axis[-1]])
        step_size.append(abs(axis[1] - axis[0]))

    range_ = np.array(range_, dtype=np.float64)
    step_size = np.array(step_size, dtype=np.float64)

    return range_, step_size


def step_size_to_num(range_, step_size):
    """Returns the number of points an axis with a range_ and step_size.

    Args:
        range_ (list): Max and min value of the axis.
        step_size (float): Step size of the axis.

    Returns:
        (int): Number of points in the axis.
    """

    num = int((range_[1] - range_[0]) / step_size) + 1

    return num


def get_rotation_axes(phi, theta):
    """Returns the rotation axes when rotating with Euler angles.

    Args:
        phi (float): Phi (first rotation) angle in degrees.
        theta (float): Theta (second rotation) angle in degrees.

    Returns:
        (list): List of axes (3D list). First is always unchanged
        z-axis. Second is the new x'-axis and third new z''-axis. 
    """

    cos_phi = np.cos(np.radians(phi))
    sin_phi = np.sin(np.radians(phi))
    cos_theta = np.cos(np.radians(theta))
    sin_theta = np.sin(np.radians(theta))

    # First axis is always the z-axis
    axis1 = [0, 0, 1]
    # x'-axis
    axis2 = [cos_phi, sin_phi, 0]
    # z''-axis
    axis3 = [sin_phi * sin_theta, -cos_phi *
             sin_theta, cos_theta]

    return [axis1, axis2, axis3]


def energy_to_k(E_kin):
    """Convert kinetic energy in [eV] to k in [Anstroem^-1].

    Args:
        E_kin (float): Kinetic energy in [eV].

    Returns:
        (float): Wave number in [Anstroem^-1].
    """

    # Electron mass
    m = 9.10938356e-31
    # Reduced Planck constant
    hbar = 1.0545718e-34
    # Electron charge
    e = 1.60217662e-19

    fac = 2 * 1e-20 * e * m / hbar**2

    return np.sqrt(fac * E_kin)


def compute_Euler_matrix(phi, theta, psi):
    """Compute rotation matrix according to Eq. (M3.10.3) of Lang-Pucker.

    Args:
        phi (float): Phi (first rotation) angle in degrees.
        theta (float): Theta (second rotation) angle in degrees.
        psi (float): Psi (third rotation) angle in degrees.

    Returns:
        (array): 3x3 Rotation matrix.
    """

    # Compute sines and cosines of angles
    sin_phi = np.sin(np.radians(phi))
    cos_phi = np.cos(np.radians(phi))
    sin_theta = np.sin(np.radians(theta))
    cos_theta = np.cos(np.radians(theta))
    sin_psi = np.sin(np.radians(psi))
    cos_psi = np.cos(np.radians(psi))

    # Euler matrix r according to eq. (M3.10.3) of Lang-Pucker
    r = np.zeros((3, 3))
    r[0, 0] = cos_phi * cos_psi - sin_phi * cos_theta * sin_psi
    r[1, 0] = -cos_phi * sin_psi - sin_phi * cos_theta * cos_psi
    r[2, 0] = sin_phi * sin_theta

    r[0, 1] = sin_phi * cos_psi + cos_phi * cos_theta * sin_psi
    r[1, 1] = -sin_phi * sin_psi + cos_phi * cos_theta * cos_psi
    r[2, 1] = -cos_phi * sin_theta

    r[0, 2] = sin_theta * sin_psi
    r[1, 2] = sin_theta * cos_psi
    r[2, 2] = cos_theta

    return r


def get_reduced_chi2(data, n):
    """Returns the reduced Chi^2 (sum of squares normalized by number of
        non NaN values minus the degrees of freedom 'n').

    Args:
        data (list): List of arbitrary shape. Contains the NOT
            squared data.
        n (int): Degrees of freedom.

    Returns:
        (float): Reduced Chi^2 value.
    """

    N = np.count_nonzero(~np.isnan(data))
    reduced_chi2 = np.nansum(np.array(data)**2) / (N - n)

    return reduced_chi2


def transpose_axis_order(constant_axis):
    if constant_axis == 0:
        axis_order = (0, 2, 1)

    elif constant_axis == 1:
        axis_order = (2, 1, 0)

    else:
        axis_order = (1, 0, 2)

    return axis_order


def split_view(data_1, data_2, type_, scale=1):
    data = data_1.copy()

    if type_ == 'Top Bottom':
        half_idx = int(np.ceil(data.data.shape[0] / 2))
        data.data[:half_idx] = scale * data_2.data[:half_idx]

    elif type_ == 'Bottom Top':
        half_idx = int(np.ceil(data.data.shape[0] / 2))
        data.data[half_idx:] = scale * data_2.data[half_idx:]

    elif type_ == 'Right Left':
        half_idx = int(np.ceil(data.data.shape[1] / 2))
        data.data.T[:half_idx] = scale * data_2.data.T[:half_idx]

    elif type_ == 'Left Right':
        half_idx = int(np.ceil(data.data.shape[1] / 2))
        data.data.T[half_idx:] = scale * data_2.data.T[half_idx:]

    return data


import os
from pathlib import Path
from configparser import ConfigParser

from kmap import __directory__


class Config:

    def __init__(self):

        self.path_settings_default = __directory__ / 'config/settings_default.ini'
        self.path_settings_user = __directory__ / 'config/settings_user.ini'
        self.path_logging_default = __directory__ / 'config/logging_default.ini'
        self.path_logging_user = __directory__ / 'config/logging_user.ini'
        self.path_shortcut_default = __directory__ / 'config/shortcut_default.ini'
        self.path_shortcut_user = __directory__ / 'config/shortcut_user.ini'

        self._general_settings = ''
        self._logging_settings = ''
        self._shortcut_settings = ''

        self.setup()

    def setup(self):

        self._check_for_user_files()

        self._general_settings = ConfigParser()
        self._general_settings.read(
            [self.path_settings_default, self.path_settings_user])

        self._logging_settings = ConfigParser()
        self._logging_settings.read(
            [self.path_logging_default, self.path_logging_user])

        self._shortcut_settings = ConfigParser()
        self._shortcut_settings.read(
            [self.path_shortcut_default, self.path_shortcut_user])

    def get_config(self, file='general'):

        if file == 'general':
            return self._general_settings

        elif file == 'logging':
            return self._logging_settings

        elif file == 'shortcut':
            return self._shortcut_settings

        else:
            raise NotImplementedError('%s not implemented' % file)

    def get_key(self, group, key, file='general'):

        if file == 'general':
            return self._general_settings[group][key]

        elif file == 'logging':
            return self._logging_settings[group][key]

        elif file == 'shortcut':
            return self._shortcut_settings[group][key]

        else:
            raise NotImplementedError('%s not implemented' % file)

    def set_key(self, group, key, value, file='general'):
        pass

    def _check_for_user_files(self):

        if not os.path.isfile(self.path_settings_user):
            with open(self.path_settings_user, 'w+') as file:
                file.write('; settings_user.ini')

        if not os.path.isfile(self.path_logging_user):
            with open(self.path_logging_user, 'w+') as file:
                file.write('; logging_user.ini')

        if not os.path.isfile(self.path_shortcut_user):
            with open(self.path_shortcut_user, 'w+') as file:
                file.write('; shortcut_user.ini')


config = Config()



# Python Imports
import logging

# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QWidget

# Own Imports
from kmap import __directory__

# Load .ui File
UI_file = __directory__ / 'ui/cubeoptions.ui'
CubeOptions_UI, _ = uic.loadUiType(UI_file)


class CubeOptions(QWidget, CubeOptions_UI):
    energy_changed = pyqtSignal()
    resolution_changed = pyqtSignal()
    symmetrization_changed = pyqtSignal()
    get_match_energy = pyqtSignal()

    def __init__(self, *args, **kwargs):
        # Setup GUI
        super(CubeOptions, self).__init__(*args, **kwargs)
        self.setupUi(self)
        self._connect()

    def save_state(self):
        E_kin = self.energy_spinbox.value()
        resolution = self.resolution_spinbox.value()
        symmetry = self.symmetrize_combobox.currentIndex()

        save = {'E_kin': E_kin, 'resolution': resolution, 'symmetry': symmetry}

        return save

    def restore_state(self, save):
        E_kin = save['E_kin']
        resolution = save['resolution']
        symmetry = save['symmetry']

        self.energy_spinbox.setValue(E_kin)
        self.resolution_spinbox.setValue(resolution)
        self.symmetrize_combobox.setCurrentIndex(symmetry)

    def set_kinetic_energy(self, energy):
        if energy is not None:
            self.energy_spinbox.setValue(energy)

    def get_parameters(self):
        energy = self.energy_spinbox.value()
        resolution = self.resolution_spinbox.value()
        symmetry = self._get_symmetry()

        return energy, resolution, symmetry

    def _get_symmetry(self):
        index = self.symmetrize_combobox.currentIndex()
        available_symmetries = ['no', '2-fold', '2-fold+mirror',
                                '3-fold', '3-fold+mirror', '4-fold',
                                '4-fold+mirror']

        return available_symmetries[index]

    def _connect(self):
        self.energy_spinbox.valueChanged.connect(self.energy_changed.emit)
        self.resolution_spinbox.valueChanged.connect(
            self.resolution_changed.emit)
        self.symmetrize_combobox.currentIndexChanged.connect(
            self.symmetrization_changed.emit)
        self.match_button.clicked.connect(self.get_match_energy.emit)



from kmap.library.id import ID
from kmap.library.sliceddata import SlicedData
from kmap.library.misc import transpose_axis_order
from kmap.config.config import config


class SlicedDataTabModel():

    def __init__(self):
        self.data = None
        self.load_data = None
        self.symmetry = ['no', False]

    def load_data_from_URLs(self, URLs):  # -> create data[BE,kx,ky]

        # Last element in URLs are the parameters. All other elements
        # are individual orbitals to load. Each is a list of length 2
        # with first entry being the URL, the second the meta_data
        # dictionary
        *orbitals, options = URLs
        name, *parameters = options
        s_share = float(config.get_key('orbital', 's_share'))

        self.data = SlicedData.init_from_orbitals(name, orbitals,
                                                  parameters, s_share=s_share)
        self.load_data = ['load_from_URLs', URLs]

        self.change_slice(0, 0)

    def load_data_from_URL(self, URL):  # -> create data[photon_energy,kx,ky]

        # Last element in URL are the parameters. The first element
        # is the orbital to load as a list of length 2
        # with first entry being the URL, the second the meta_data
        # dictionary
        *orbital, options = URL
        name, *parameters = options
        s_share = float(config.get_key('orbital', 's_share'))

        self.data = SlicedData.init_from_orbital_photonenergy(
            name, orbital, parameters, s_share=s_share)
        self.load_data = ['load_from_URL', URL]

        self.change_slice(0, 0)

    def load_data_from_cube(self, URL):  # -> create either psi[x,y,z]
        #               or psik[kx,ky,kz]

        # Last element in URL are the parameters. The first element
        # is the orbital to load as a list of length 2
        # with first entry being the URL, the second the meta_data
        # dictionary
        *orbital, options = URL
        name, *parameters = options
        self.data = SlicedData.init_from_orbital_cube(
            name, orbital, parameters)
        self.load_data = ['load_from_cube', URL]

        self.change_slice(0, 0)

    def load_data_from_path(self, path):
        self.data = SlicedData.init_from_hdf5(path)
        self.load_data = ['load_from_path', path]

        self.change_slice(0, 0)

    def save_state(self):
        load_type, load_args = self.load_data
        save = {'load_type': load_type,
                'load_args': load_args, 'ID': self.data.ID}

        return save

    def transpose(self, constant_axis):
        axis_order = transpose_axis_order(constant_axis)

        self.data.transpose(axis_order)

    def change_symmetry(self, symmetry, mirror):
        self.symmetry = [symmetry, mirror]

    def change_slice(self, index, axis):
        self.displayed_plot_data = self.data.slice_from_index(index, axis)
        self.displayed_plot_data.symmetrise(*self.symmetry)

        return self.displayed_plot_data

    def to_string(self):
        rep = '%s' % str(self.data)

        return rep



# Python Imports
import logging

# Third Party Imports
import h5py
import numpy as np

# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QFileDialog

# Own Imports
from kmap import __directory__
from kmap.library.qwidgetsub import Tab
from kmap.model.orbitaldatatab_model import OrbitalDataTabModel
from kmap.controller.matplotlibwindow import MatplotlibImageWindow
from kmap.controller.crosshairannulus import CrosshairAnnulus
from kmap.controller.colormap import Colormap
from kmap.controller.cubeoptions import CubeOptions
from kmap.controller.orbitaltable import OrbitalTable
from kmap.controller.interpolation import Interpolation
from kmap.controller.pyqtgraphplot import PyQtGraphPlot
from kmap.controller.polarization import Polarization
from kmap.library.sliceddata import Axis
from kmap.config.config import config

# Load .ui File
UI_file = __directory__ / 'ui/orbitaldatatab.ui'
OrbitalDataTab_UI, _ = uic.loadUiType(UI_file)


class OrbitalDataTab(Tab, OrbitalDataTab_UI):
    orbital_removed = pyqtSignal(int)
    orbital_added = pyqtSignal(int)
    get_energy = pyqtSignal()

    def __init__(self):
        # Setup GUI
        super(OrbitalDataTab, self).__init__()
        self.setupUi(self)
        self._setup()
        self._connect()

        self.model = OrbitalDataTabModel(self)

    def add_orbital_from_filepath(self, path):
        orbital = self.model.load_data_from_path(path)
        self.add_orbital(orbital)

        return orbital.ID

    def add_orbital_from_online(self, URL, meta_data={}):
        orbital = self.model.load_data_from_online(URL, meta_data)
        self.add_orbital(orbital)

        return orbital.ID

    def get_orbitals(self):
        orbitals = [orbital[0] for orbital in self.model.orbitals]

        return orbitals

    def add_orbital(self, orbital):
        if 'orientation' in orbital.meta_data:
            orientation = orbital.meta_data['orientation']

        else:
            orientation = 'xy'

        self.table.add_orbital(orbital, orientation)

        self.refresh_plot()

        self.orbital_added.emit(orbital.ID)

    def refresh_plot(self):
        data = self.model.update_displayed_plot_data()

        data = self.interpolation.interpolate(data)
        data = self.interpolation.smooth(data)

        self.plot_item.plot(data)

    def refresh_mini_plots(self, ID, orbital_changed=True):
        parameters = self.get_parameters(ID)
        phi, theta, psi = parameters[3:6]
        E_kin = parameters[1]

        if orbital_changed:
            orbital = self.model.ID_to_orbital(ID)
            self.mini_3Dkspace_plot.set_orbital(orbital, ID)
            self.mini_real_plot.set_orbital(orbital)

        self.mini_3Dkspace_plot.change_energy(E_kin)
        self.mini_3Dkspace_plot.rotate_orbital(phi, theta, psi)
        self.mini_real_plot.rotate_orbital(phi, theta, psi)

    def refresh_mini_plot_polarization(self):
        polarization, alpha, beta = self.polarization.get_parameters()[1:4]

        self.mini_real_plot.rotate_photon(polarization, alpha, beta)

    def get_crosshair(self):
        return self.crosshair

    def get_parameters(self, ID):
        kinetic_energy, dk, symmetry = self.cube_options.get_parameters()
        parameters = self.table.get_parameters_by_ID(ID)
        weight, *orientation = parameters
        polarization = self.polarization.get_parameters()

        return (weight, kinetic_energy, dk,
                *orientation, *polarization, symmetry)

    def get_use(self, ID):
        return self.table.get_use_by_ID(ID)

    def get_displayed_plot_data(self):
        return self.model.displayed_plot_data

    def crosshair_changed(self):
        self.crosshair.update_label()

    def change_parameter(self):
        self.refresh_plot()

        ID = self.mini_3Dkspace_plot.ID
        if ID is not None:
            self.refresh_mini_plots(ID, orbital_changed=False)

    def orbitals_changed(self, ID):
        self.refresh_plot()

        current_ID = self.mini_3Dkspace_plot.ID
        if current_ID is None or current_ID == ID:
            self.refresh_mini_plots(ID, orbital_changed=False)

    def remove_orbital_by_ID(self, ID):
        self.model.remove_orbital_by_ID(ID)

        self.refresh_plot()
        self.mini_3Dkspace_plot.set_orbital(None, ID)
        self.mini_real_plot.set_orbital(None)

        self.orbital_removed.emit(ID)

    def export_to_hdf5(self):
        if not self.interpolation.interpolation_checkbox.isChecked():
            print('Only interpolated OrbitalData can be exported.')
            return

        path = config.get_key('paths', 'hdf5_export_start')
        if path == 'None':
            file_name, _ = QFileDialog.getSaveFileName(
                None, 'Save .hdf5 File (*.hdf5)')
        else:
            start_path = str(__directory__ / path)
            file_name, _ = QFileDialog.getSaveFileName(
                None, 'Save .hdf5 File (*.hdf5)', str(start_path))

        if not file_name:
            return
        else:
            h5file = h5py.File(file_name, 'w')

        export_energies = eval(config.get_key('orbital', 'export_energies'))
        if isinstance(export_energies, dict):
            export_energies = np.linspace(export_energies['min'],
                                          export_energies['max'],
                                          export_energies['num'],
                                          endpoint=True)

        kmaps = []
        old_energy = self.cube_options.energy_spinbox.value()
        for energy in export_energies:
            self.cube_options.energy_spinbox.setValue(energy)
            kmaps.append(self.get_displayed_plot_data().data)
        self.cube_options.energy_spinbox.setValue(old_energy)
        kmaps = np.array(kmaps)
        xrange, yrange = self.get_displayed_plot_data().range

        h5file.create_dataset('name', data='Orbitals')
        h5file.create_dataset('axis_1_label', data='E_kin')
        h5file.create_dataset('axis_2_label', data='kx')
        h5file.create_dataset('axis_3_label', data='ky')
        h5file.create_dataset('axis_1_units', data='eV')
        h5file.create_dataset('axis_2_units', data='1/Å')
        h5file.create_dataset('axis_3_units', data='1/Å')
        h5file.create_dataset('axis_1_range', data=[export_energies[0],
                                                    export_energies[-1]])
        h5file.create_dataset('axis_2_range', data=xrange)
        h5file.create_dataset('axis_3_range', data=yrange)
        h5file.create_dataset('data', data=kmaps, dtype='f8',
                              compression='gzip', compression_opts=9)
        h5file.close()

    def display_in_matplotlib(self):
        data = self.model.displayed_plot_data
        LUT = self.plot_item.get_LUT()

        window = MatplotlibImageWindow(data, LUT=LUT)

        return window

    def change_kinetic_energy(self, energy):
        self.cube_options.set_kinetic_energy(energy)

    def closeEvent(self, event):
        del self.model

        Tab.closeEvent(self, event)

    def get_plot_labels(self):
        bottom = self.plot_item.get_label('bottom')
        left = self.plot_item.get_label('left')
        return bottom, left

    def save_state(self):
        crosshair_save = self.crosshair.save_state()
        interpolation_save = self.interpolation.save_state()
        polarization_save = self.polarization.save_state()
        cube_options_save = self.cube_options.save_state()
        real_space_options_save = self.real_space_options.save_state()
        colormap_save = self.colormap.save_state()

        orbital_save = []
        for orbital in self.model.orbitals:
            parameters = self.get_parameters(orbital[0].ID)
            use = self.get_use(orbital[0].ID)
            orbital_save.append([*orbital[1:], parameters, use])

        save = {'crosshair': crosshair_save,
                'interpolation': interpolation_save,
                'polarization': polarization_save,
                'cube_options': cube_options_save,
                'real_space_options': real_space_options_save,
                'orbital': orbital_save,
                'colormap': colormap_save,
                'title': self.title}

        return save, []

    def restore_state(self, save):
        ID_maps = []
        for orbital in save['orbital']:
            if orbital[0] == 'path':
                ID = self.add_orbital_from_filepath(orbital[1])

            elif orbital[0] == 'url':
                ID = self.add_orbital_from_online(orbital[1], orbital[2])

            else:
                raise ValueError

            ID_maps.append([orbital[3], ID])
            parameter = [orbital[4][0], *orbital[4][3:6]]
            use = orbital[5]
            self.table.update_orbital_parameters(ID, parameter)
            self.table.update_orbital_use(ID, use)

        crosshair_save = save['crosshair']
        interpolation_save = save['interpolation']
        polarization_save = save['polarization']
        cube_options_save = save['cube_options']
        real_space_options_save = save['real_space_options']
        colormap_save = save['colormap']

        self.crosshair.restore_state(crosshair_save)
        self.interpolation.restore_state(interpolation_save)
        self.polarization.restore_state(polarization_save)
        self.cube_options.restore_state(cube_options_save)
        self.real_space_options.restore_state(real_space_options_save)
        self.colormap.restore_state(colormap_save)

        return ID_maps

    def _setup(self):
        self.crosshair = CrosshairAnnulus(self.plot_item)
        self.colormap = Colormap([self.plot_item])
        self.interpolation = Interpolation()

        layout = self.scroll_area.widget().layout()
        layout.insertWidget(4, self.interpolation)
        layout.insertWidget(3, self.colormap)
        layout.insertWidget(5, self.crosshair)

        self.mini_real_plot.set_options(self.real_space_options)
        self.mini_3Dkspace_plot.set_options(self.real_space_options)

        # Rough axis values for all orbitals to set labels for interpolation
        x = Axis('kx', '1/Å', [-3, 3], 200)
        y = Axis('ky', '1/Å', [-3, 3], 200)
        self.interpolation.set_label(x, y)
        self.plot_item.set_labels(x, y)

        self.title = 'Orbitals'

    def _connect(self):
        self.crosshair.crosshair_changed.connect(self.crosshair_changed)
        self.table.orbital_changed.connect(self.orbitals_changed)
        self.table.orbital_removed.connect(self.remove_orbital_by_ID)
        self.table.orbital_selected.connect(self.refresh_mini_plots)
        self.polarization.polarization_changed.connect(
            self.change_parameter)
        self.polarization.polarization_changed.connect(
            self.refresh_mini_plot_polarization)
        self.cube_options.symmetrization_changed.connect(
            self.change_parameter)
        self.cube_options.energy_changed.connect(
            self.change_parameter)
        self.cube_options.resolution_changed.connect(
            self.change_parameter)
        self.cube_options.get_match_energy.connect(self.get_energy.emit)

        self.interpolation.smoothing_changed.connect(self.refresh_plot)
        self.interpolation.interpolation_changed.connect(self.refresh_plot)



import os
import urllib.request
from kmap.library.orbital import Orbital
from kmap.config.config import config
from kmap.library.abstractdata import AbstractData


class OrbitalData(Orbital, AbstractData):

    def __init__(self, cube, ID, name='', meta_data={}):

        self.dk3D = float(config.get_key('orbital', 'dk3D'))

        AbstractData.__init__(self, ID, name, meta_data)
        Orbital.__init__(self, cube, file_format='cube', dk3D=self.dk3D,
                         value='abs2')

    @classmethod
    def init_from_file(cls, file_path, ID):
        with open(file_path, 'r') as f:
            file = f.read()

            name, keys = OrbitalData._get_metadata(file, file_path)

        return cls(file, ID, name=name, meta_data=keys)

    @classmethod
    def init_from_online(cls, url, ID, meta_data={}):

        with urllib.request.urlopen(url) as f:
            file = f.read().decode('utf-8')

            name, keys = OrbitalData._get_metadata(file, url)
            name = meta_data['name'] if 'name' in meta_data else name
            meta_data.update(keys)
            
              

        return cls(file, ID, name=name, meta_data=meta_data)

    @classmethod
    def _get_metadata(cls, file, file_path):

        first_line, second_line = file.split('\n')[:2]

        name = os.path.splitext(os.path.split(file_path)[1])[0]
        keys = {config.get_key('cube', 'line_one'): first_line.strip(),
                config.get_key('cube', 'line_two'): second_line.strip()}

        return name, keys

    def __str__(self):

        rep = AbstractData.__str__(self)
        rep += '\ndk3D:\t\t%s' % self.dk3D

        return rep



# Third Party Imports
import numpy as np

# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QWidget

# Own Imports
from kmap import __directory__
from kmap.config.config import config


class DataSliderBase(QWidget):
    slice_changed = pyqtSignal(int)
    axis_changed = pyqtSignal(int)
    symmetry_changed = pyqtSignal(str, bool)

    def __init__(self, data, *args, **kwargs):
        self.data = data

        # Setup GUI
        super(DataSliderBase, self).__init__(*args, **kwargs)

    def change_slice(self, index):
        sender = self.sender().objectName()

        # If spinbox or slider called, index will be the slice index
        # and the other widget has to be updated. As default use slider
        # position as index
        if sender == 'spinbox':
            self._update_slider_silently(index)

        elif sender == 'slider':
            self._update_spinbox_silently(index)

        else:
            index = self.slider.sliderPosition()

        self._update_slice_label()

        self.slice_changed.emit(index)

    def change_axis(self, axis):
        index = self.slider.sliderPosition()

        self._update_slider_silently(index)
        self._update_spinbox_silently(index)
        self._update_slice_label()

        self.axis_changed.emit(axis)

    def change_symmetry(self, index):
        symmetries = ('no', '2-fold', '3-fold', '4-fold')
        symmetry = symmetries[int(np.ceil(index / 2))]
        mirror = True if index in (2, 4, 6) else False
        self.symmetry_changed.emit(symmetry, mirror)

    def get_index(self):
        index = self.slider.sliderPosition()

        return index

    def get_axis(self):
        axis = self.combobox.currentIndex()

        return axis

    def save_state(self):
        slice_ = self.get_index()
        axis = self.get_axis()

        save = {'slice': slice_, 'axis': axis}

        return save

    def restore_state(self, save):
        self.combobox.setCurrentIndex(save['axis'])
        self.slider.setValue(save['slice'])

    def _update_slice_label(self):
        index = self.slider.sliderPosition()
        axis = self.data.axes[self.combobox.currentIndex()]
        value = axis.axis[index]

        text = '%.2f  %s' % (value, axis.units)
        self.value_label.setText(text)

    def _update_slider_silently(self, index):
        self.slider.blockSignals(True)

        axis = self.data.axes[self.combobox.currentIndex()]
        maximum = len(axis.axis) - 1
        self.slider.setMaximum(maximum)
        self.slider.setSliderPosition(index)

        self.slider.blockSignals(False)

    def _update_spinbox_silently(self, index):
        self.spinbox.blockSignals(True)

        axis = self.data.axes[self.combobox.currentIndex()]
        maximum = len(axis.axis) - 1
        self.spinbox.setMaximum(maximum)
        self.spinbox.setValue(index)

        self.spinbox.blockSignals(False)

    def _setup_combobox(self):
        for index in range(3):
            axis_label = self.data.axes[index].label
            self.combobox.addItem(str(axis_label))

    def _setup(self):
        self._setup_combobox()
        self._update_slice_label()
        self.change_axis(0)

    def _connect(self):
        self.slider.valueChanged.connect(self.change_slice)
        self.spinbox.valueChanged.connect(self.change_slice)
        self.combobox.currentIndexChanged.connect(self.change_axis)
        self.symmetrize_combobox.currentIndexChanged.connect(
            self.change_symmetry)


# Load .ui File
UI_file = __directory__ / 'ui/dataslidernotranspose.ui'
DataSliderNoTranspose_UI, _ = uic.loadUiType(UI_file)


class DataSliderNoTranspose(DataSliderBase, DataSliderNoTranspose_UI):
    def __init__(self, *args, **kwargs):
        # Setup GUI
        super(DataSliderNoTranspose, self).__init__(*args, **kwargs)
        self.setupUi(self)
        self._setup()
        self._connect()


# Load .ui File
UI_file = __directory__ / 'ui/dataslider.ui'
DataSlider_UI, _ = uic.loadUiType(UI_file)


class DataSlider(DataSliderBase, DataSlider_UI):
    tranpose_triggered = pyqtSignal(int)
    symmetry_changed = pyqtSignal(str, bool)

    def __init__(self, *args, **kwargs):
        # Setup GUI
        super(DataSlider, self).__init__(*args, **kwargs)
        self.setupUi(self)
        self._setup()
        self._connect()

    def trigger_transpose(self):
        index = self.slider.sliderPosition()

        self.tranpose_triggered.emit(self.get_axis())

        for i in range(3):
            self.combobox.setItemText(i, self.data.axes[i].label)

        self._update_slice_label()
        self._update_slider_silently(index)
        self._update_spinbox_silently(index)

    def _connect(self):
        super()._connect()
        self.transpose_button.clicked.connect(self.trigger_transpose)



# Python Imports
import os
import logging
import traceback
from shutil import copy

# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtWidgets import QWidget

# Own Imports
from kmap import __directory__
from kmap.model.colormap_model import ColormapModel
from kmap.config.config import config

# Load .ui File
UI_file = __directory__ / 'ui/colormap.ui'
Colormap_UI, _ = uic.loadUiType(UI_file)


class Colormap(QWidget, Colormap_UI):

    def __init__(self, plot_item):

        # Setup GUI
        super(Colormap, self).__init__()
        self.setupUi(self)
        self._connect()

        self.path = self.get_path()
        self.model = ColormapModel(plot_item)

        self.load_colormaps()

    def load_colormaps(self):

        try:
            self.model.load_colormaps(self.path)

        except Exception:

            log = logging.getLogger('kmap')

            log.error('Colormaps could not be loaded')
            log.error(traceback.format_exc())

        self._update_combobox()
        self.set_default_colormap()

    def save_state(self):

        current_colormap = self.combobox.currentText()

        save = {'current_colormap': current_colormap}

        return save

    def get_path(self):

        # Path for the .json file containing the colormaps
        temp = __directory__ / config.get_key('paths', 'colormap')
        user = temp / 'colormaps_user.json'

        if not os.path.isfile(user):
            default = temp / 'colormaps_default.json'
            copy(default, user)

        return user
        

    def restore_state(self, save):

        try:
            self.set_colormap(save['current_colormap'])

        except ValueError:
            log = logging.getLogger('kmap')

            log.error(
                'This colormap does not exist. Please save the colormap first.')
            log.error(traceback.format_exc())

    def add_colormap(self):

        name = self.line_edit.text()

        if name:
            self.model.add_colormap_from_plot(name)
            self._update_combobox()
            self.set_colormap(name)

    def remove_colormap(self):

        name = self.combobox.currentText()

        if name:
            self.model.remove_colormap(name)
            self._update_combobox()
            self.set_default_colormap()

    def save(self):

        self.model.save_colormaps(self.path)

    def set_colormap(self, name):

        self.model.set_current_colormap(name)
        self.combobox.setCurrentText(name)

    def set_default_colormap(self):

        default_colormap = config.get_key('colormap', 'default')

        try:
            self.set_colormap(default_colormap)

        except ValueError:
            log = logging.getLogger('kmap')
            log.error(
                'The colormap set as default does not exist.')
            log.error(traceback.format_exc())

    def _update_combobox(self):

        self.combobox.blockSignals(True)

        self.combobox.clear()
        for colormap in self.model.colormaps:
            self.combobox.addItem(colormap.name)

        self.combobox.blockSignals(False)

    def _connect(self):

        self.combobox.currentTextChanged.connect(self.set_colormap)
        self.remove_button.clicked.connect(self.remove_colormap)
        self.add_button.clicked.connect(self.add_colormap)
        self.save_button.clicked.connect(self.save)
        self.reload_button.clicked.connect(self.load_colormaps)



import numpy as np
from kmap.library.plotdata import PlotData
from kmap.library.misc import split_view


class SplitViewTabModel():

    def __init__(self, sliced_data_tab, orbital_data_tab, interpolation):
        self.sliced_data_tab = sliced_data_tab
        self.orbital_data_tab = orbital_data_tab
        self.interpolation = interpolation
        self.displayed_sliced_data = None
        self.displayed_plot_data = None
        self.symmetry = ['no', False]
        self.scale = 1
        self.split_type = 'Left Right'

        self.update_displayed_plot_data(0, 0)

    def set_scale(self, scale):
        self.scale = scale

    def set_type(self, type_):
        self.split_type = type_

    def change_symmetry(self, symmetry, mirror):
        self.symmetry = [symmetry, mirror]

    def get_sliced_data(self):
        return self.sliced_data_tab.get_data()

    def update_displayed_plot_data(self, index, axis):
        temp = self.get_sliced_data().slice_from_index(
            index, axis).copy()
        temp = self.interpolation.interpolate(temp)
        sliced_data = self.interpolation.smooth(temp)
        sliced_data = sliced_data.symmetrise(*self.symmetry)

        temp = self.orbital_data_tab.get_displayed_plot_data().copy()
        temp = self.interpolation.interpolate(temp)
        orbital_data = self.interpolation.smooth(temp)

        self.displayed_plot_data = split_view(sliced_data, orbital_data,
                                              type_=self.split_type,
                                              scale=self.scale)

        return self.displayed_plot_data



# Third Party Imports
import numpy as np

# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QWidget

# Own Imports
from kmap import __directory__
from kmap.library.misc import step_size_to_num, axis_from_range


class InterpolationBase(QWidget):
    interpolation_changed = pyqtSignal()

    def __init__(self):
        # Setup GUI
        super(InterpolationBase, self).__init__()

    def interpolate(self, data):
        if self.interpolation_checkbox.isChecked():
            axes = self.get_axes()
            data.interpolate(*axes, update=True)

        return data

    def get_order(self):
        return self.order_spinbox.value()

    def get_axes(self):
        range_ = self.get_range()
        resolution = self.get_resolution()
        num = [step_size_to_num(range_[0], resolution[0]),
               step_size_to_num(range_[1], resolution[1])]

        axes = [axis_from_range(range_[0], num[0]),
                axis_from_range(range_[1], num[1])]

        return axes

    def _change_interpolation(self):
        self.interpolation_changed.emit()


# Load .ui File
UI_file = __directory__ / 'ui/interpolation.ui'
Interpolation_UI, _ = uic.loadUiType(UI_file)


class Interpolation(InterpolationBase, Interpolation_UI):
    smoothing_changed = pyqtSignal()

    def __init__(self):
        # Setup GUI
        super(Interpolation, self).__init__()
        self.setupUi(self)
        self._connect()

    def set_force_interpolation(self, bool):
        self.interpolation_checkbox.setChecked(bool)
        self.interpolation_checkbox.setEnabled(not bool)

    def smooth(self, data):
        if self.smoothing_checkbox.isChecked():
            sigma = self.get_sigma()

            if self.fill_combobox.currentIndex() == 0:
                fill_value = np.nanmean(data.data)

            elif self.fill_combobox.currentIndex() == 1:
                fill_value = 0

            else:
                fill_value = np.nan

            data.smooth(*sigma, update=True, fill_value=fill_value)

        return data

    def get_sigma(self, pixel=True):
        sigma_x = self.sigma_x_spinbox.value()
        sigma_y = self.sigma_y_spinbox.value()

        return [sigma_x, sigma_y]

    def get_range(self):
        x_min = self.x_min_spinbox.value()
        x_max = self.x_max_spinbox.value()
        y_min = self.y_min_spinbox.value()
        y_max = self.y_max_spinbox.value()

        return [[x_min, x_max], [y_min, y_max]]

    def get_resolution(self):
        x_resolution = self.x_resolution_spinbox.value()
        y_resolution = self.y_resolution_spinbox.value()

        return [x_resolution, y_resolution]

    def set_label(self, x, y):
        # Set Label
        self.x_label.setText('%s:' % x.label)
        self.y_label.setText('%s:' % y.label)

        # Set Unit
        self.x_resolution_spinbox.setSuffix('  %s' % x.units)
        self.sigma_x_spinbox.setSuffix('  %s' % x.units)
        self.y_resolution_spinbox.setSuffix('  %s' % y.units)
        self.sigma_y_spinbox.setSuffix('  %s' % y.units)
        self.x_min_spinbox.setSuffix('  %s' % x.units)
        self.x_max_spinbox.setSuffix('  %s' % x.units)
        self.y_min_spinbox.setSuffix('  %s' % y.units)
        self.y_max_spinbox.setSuffix('  %s' % y.units)

        # Set Range
        self.sigma_x_spinbox.setMinimum(0)
        self.sigma_x_spinbox.setMaximum(15 * x.stepsize)
        self.sigma_y_spinbox.setMinimum(0)
        self.sigma_y_spinbox.setMaximum(15 * y.stepsize)
        # Set min/max of min/max spinbox to twice the range
        self.x_min_spinbox.setMinimum(x.range[0] - abs(x.range[0]))
        self.y_min_spinbox.setMinimum(y.range[0] - abs(y.range[0]))
        self.x_max_spinbox.setMaximum(x.range[1] + abs(x.range[1]))
        self.y_max_spinbox.setMaximum(y.range[1] + abs(y.range[1]))

        # Set Value
        self.sigma_x_spinbox.setValue(3 * x.stepsize)
        self.x_min_spinbox.setValue(x.range[0])
        self.x_max_spinbox.setValue(x.range[-1])
        self.sigma_y_spinbox.setValue(3 * y.stepsize)
        self.y_min_spinbox.setValue(y.range[0])
        self.y_max_spinbox.setValue(y.range[-1])

        self._update_dynamic_range_spinboxes()

    def save_state(self):
        save = {'sigma': self.get_sigma(),
                'range': self.get_range(),
                'force_interpolation':
                    not self.interpolation_checkbox.isEnabled(),
                'resolution': self.get_resolution(),
                'enable_smoothing': self.smoothing_checkbox.checkState(),
                'enable_interpolation': self.interpolation_checkbox.checkState(),
                'fill_value': self.fill_combobox.currentIndex()}

        return save

    def restore_state(self, save):
        self.set_sigma(save['sigma'])
        self.set_range(save['range'])
        self.set_resolution(save['resolution'])
        self.fill_combobox.setCurrentIndex(save['fill_value'])
        self.interpolation_checkbox.setCheckState(save['enable_interpolation'])
        self.smoothing_checkbox.setCheckState(save['enable_smoothing'])
        self.set_force_interpolation(save['force_interpolation'])

    def set_sigma(self, sigma):
        self.sigma_x_spinbox.setValue(sigma[0])
        self.sigma_y_spinbox.setValue(sigma[1])

    def set_range(self, range_):
        self.x_min_spinbox.setValue(range_[0][0])
        self.x_max_spinbox.setValue(range_[0][1])
        self.y_min_spinbox.setValue(range_[1][0])
        self.y_max_spinbox.setValue(range_[1][1])

    def set_resolution(self, resolution):
        self.x_resolution_spinbox.setValue(resolution[0])
        self.y_resolution_spinbox.setValue(resolution[1])

    def _update_dynamic_range_spinboxes(self):
        # Set max/min of min/max spinbox to value of other spinbox
        self.x_min_spinbox.setMaximum(self.x_max_spinbox.value() - 1)
        self.y_min_spinbox.setMaximum(self.y_max_spinbox.value() - 1)
        self.x_max_spinbox.setMinimum(self.x_min_spinbox.value() + 1)
        self.y_max_spinbox.setMinimum(self.y_min_spinbox.value() + 1)

    def _change_smoothing(self):
        self.smoothing_changed.emit()

    def _connect(self):
        self.sigma_x_spinbox.valueChanged.connect(self._change_smoothing)
        self.sigma_y_spinbox.valueChanged.connect(self._change_smoothing)
        self.smoothing_checkbox.stateChanged.connect(
            self._change_smoothing)

        self.x_min_spinbox.valueChanged.connect(
            self._update_dynamic_range_spinboxes)
        self.x_max_spinbox.valueChanged.connect(
            self._update_dynamic_range_spinboxes)
        self.y_min_spinbox.valueChanged.connect(
            self._update_dynamic_range_spinboxes)
        self.y_max_spinbox.valueChanged.connect(
            self._update_dynamic_range_spinboxes)

        self.x_min_spinbox.valueChanged.connect(self._change_interpolation)
        self.x_max_spinbox.valueChanged.connect(self._change_interpolation)
        self.y_min_spinbox.valueChanged.connect(self._change_interpolation)
        self.y_max_spinbox.valueChanged.connect(self._change_interpolation)
        self.interpolation_checkbox.stateChanged.connect(
            self._change_interpolation)

        self.x_resolution_spinbox.valueChanged.connect(
            self._change_interpolation)
        self.y_resolution_spinbox.valueChanged.connect(
            self._change_interpolation)


# Load .ui File
UI_file = __directory__ / 'ui/lmfitinterpolation.ui'
LMFitInterpolation_UI, _ = uic.loadUiType(UI_file)


class LMFitInterpolation(InterpolationBase, LMFitInterpolation_UI):

    def __init__(self):
        # Setup GUI
        super(LMFitInterpolation, self).__init__()
        self.setupUi(self)
        self._connect()

    def save_state(self):
        save = {'range': self.get_range(),
                'resolution': self.get_resolution()}

        return save

    def restore_state(self, save):
        self.set_range(save['range'])
        self.set_resolution(save['resolution'])

    def set_range(self, range_):
        self.min_spinbox.setValue(range_[0][0])
        self.max_spinbox.setValue(range_[0][1])

    def get_range(self):
        min_ = self.min_spinbox.value()
        max_ = self.max_spinbox.value()

        return [[min_, max_], [min_, max_]]

    def set_resolution(self, resolution):
        self.resolution_spinbox.setValue(resolution[0])

    def get_resolution(self):
        resolution = self.resolution_spinbox.value()

        return [resolution, resolution]

    def set_label(self, x, y):
        # Set Label
        self.label.setText('%s:' % x.label)

        # Set Unit
        self.resolution_spinbox.setSuffix('  %s' % x.units)
        self.min_spinbox.setSuffix('  %s' % x.units)
        self.max_spinbox.setSuffix('  %s' % x.units)

        # Set min/max of min/max spinbox to twice the range
        self.min_spinbox.setMinimum(x.range[0] - abs(x.range[0]))
        self.max_spinbox.setMaximum(x.range[1] + abs(x.range[1]))

        # Set Value
        self.min_spinbox.setValue(x.range[0])
        self.max_spinbox.setValue(x.range[-1])

        self._update_dynamic_range_spinboxes()

    def get_axis(self):
        axes = self.get_axes()

        return axes[0]

    def _update_dynamic_range_spinboxes(self):
        # Set max/min of min/max spinbox to value of other spinbox
        self.min_spinbox.setMaximum(self.max_spinbox.value() - 1)
        self.max_spinbox.setMinimum(self.min_spinbox.value() + 1)

    def _connect(self):
        self.min_spinbox.valueChanged.connect(
            self._update_dynamic_range_spinboxes)
        self.max_spinbox.valueChanged.connect(
            self._update_dynamic_range_spinboxes)

        self.min_spinbox.valueChanged.connect(self._change_interpolation)
        self.max_spinbox.valueChanged.connect(self._change_interpolation)

        self.resolution_spinbox.valueChanged.connect(
            self._change_interpolation)



# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QWidget

# Own Imports
from kmap import __directory__
from kmap.config.config import config

# Load .ui File
UI_file = __directory__ / 'ui/lmfitorbitaloptions.ui'
LMFitOrbitalOptions_UI, _ = uic.loadUiType(UI_file)


class LMFitOrbitalOptions(QWidget, LMFitOrbitalOptions_UI):
    symmetrization_changed = pyqtSignal(str)
    polarization_changed = pyqtSignal(str, str)

    def __init__(self, *args, **kwargs):
        # Setup GUI
        super(LMFitOrbitalOptions, self).__init__(*args, **kwargs)
        self.setupUi(self)
        self._connect()

    def save_state(self):
        save = {'factor': self.ak_combobox.currentIndex(),
                'polarization': self.polarization_combobox.currentIndex(),
                'symmetry': self.symmetrize_combobox.currentIndex()}

        return save

    def restore_state(self, save):
        self.ak_combobox.setCurrentIndex(save['factor'])
        self.polarization_combobox.setCurrentIndex(save['polarization'])
        self.symmetrize_combobox.setCurrentIndex(save['symmetry'])

    def get_symmetrization(self):
        index = self.symmetrize_combobox.currentIndex()
        available_symmetries = ['no', '2-fold', '2-fold+mirror',
                                '3-fold', '3-fold+mirror', '4-fold',
                                '4-fold+mirror']

        return available_symmetries[index]

    def get_polarization(self):
        Ak_index = self.polarization_combobox.currentIndex()

        if Ak_index == 0:
            Ak_type = 'toroid'
            polarization = 'p'

        else:
            Ak_type = 'NanoESCA'

            if Ak_index == 1:
                polarization = 'p'

            elif Ak_index == 2:
                polarization = 's'

            elif Ak_index == 3:
                polarization = 'unpolarized'

            elif Ak_index == 4:
                polarization = 'C+'

            elif Ak_index == 5:
                polarization = 'C-'

            elif Ak_index == 6:
                polarization = 'CDAD'

        factor = self._get_factor()
        Ak_type = factor if factor == 'no' else factor + Ak_type

        return Ak_type, polarization

    def _get_factor(self):
        Ak_index = self.ak_combobox.currentIndex()

        if Ak_index == 0:
            Ak_type = 'no'

        elif Ak_index == 1:
            Ak_type = 'only-'

        else:
            Ak_type = ''

        return Ak_type

    def _change_symmetrization(self):
        symmetrization = self.get_symmetrization()

        self.symmetrization_changed.emit(symmetrization)

    def _change_polarization(self):
        Ak_type, polarization = self.get_polarization()
        self.polarization_changed.emit(Ak_type, polarization)

    def _connect(self):
        self.polarization_combobox.currentIndexChanged.connect(
            self._change_polarization)
        self.ak_combobox.currentIndexChanged.connect(self._change_polarization)
        self.symmetrize_combobox.currentIndexChanged.connect(
            self._change_symmetrization)



# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtWidgets import QWidget

# Own Imports
from kmap import __directory__
from kmap.library.qwidgetsub import Tab
from kmap.controller.matplotlibwindow import MatplotlibLineWindow

# Load .ui File
UI_file = __directory__ / 'ui/lmfitplottab.ui'
LMFitPlotTab_UI, _ = uic.loadUiType(UI_file)


class LMFitPlotTab(Tab, LMFitPlotTab_UI):

    def __init__(self, results, orbitals, axis, residuals,
                 background_parameters, result_tab, *args,
                 **kwargs):
        self.results = results
        self.result_tab = result_tab
        self.orbitals = orbitals
        self.x_axis = axis
        self.residuals = residuals
        self.background_parameters = background_parameters

        # Setup GUI
        super(LMFitPlotTab, self).__init__(*args, **kwargs)
        self.setupUi(self)

        self._setup()
        self._connect()

        self.refresh_plot()

    @classmethod
    def init_from_save(cls, save, tab):
        results = save['results']
        axis = save['axis']
        residuals = save['residuals'] if 'residuals' in save else None
        orbitals = tab.get_orbitals()

        tab = LMFitPlotTab(results, orbitals, axis, residuals, tab)

        return tab

    def save_state(self):
        save = {'title': self.title,
                'results': self.results,
                'axis': self.x_axis,
                'residuals': self.residuals}

        return save, [self.result_tab]

    def export_to_txt(self):
        data = self.plot_item.get_data()

        text = ''

        for data_set in data:
            name = data_set['name']
            x = data_set['x']
            y = data_set['y']
            text += '# ' + name + '\n'
            for xi, yi in zip(x, y):
                text += '%g  %g \n' % (xi, yi)
            text += '\n'

        return text

    def refresh_plot(self):
        x = self.x_axis.axis
        x_label = '%s [%s]' % (self.x_axis.label, self.x_axis.units)
        self.plot_item.clear()

        option = self.parameter_combobox.currentIndex()
        if option == 4:
            # Residual
            title = 'Residual'
            y = self.residuals
            self.plot_item.plot(x, y, title)
            y_label = '|Residual|'

        elif option in list(range(4)):
            # Orbital parameters
            possible_params = ['w_', 'phi_', 'theta_', 'psi_']
            possible_labels = ['Weight [1]', 'Phi [°]', 'Theta [°]', 'Psi [°]']
            param = possible_params[option]
            y_label = possible_labels[option]

            for orbital in self.orbitals:
                y = [result.params[param + str(orbital.ID)]
                     for result in self.results]
                title = orbital.name

                self.plot_item.plot(x, y, title)

        else:
            # Background parameters
            parameter = self.background_parameters[option - 5]
            y_label = f'{parameter} [a.U.]'
            y = [result.params[parameter] for result in self.results]
            title = parameter

            self.plot_item.plot(x, y, title)

        self.plot_item.set_label(x_label, y_label)

    def display_in_matplotlib(self):
        data = self.plot_item.get_data()

        window = MatplotlibLineWindow(data)

        return window

    def _setup(self):
        for parameter in self.background_parameters:
            self.parameter_combobox.addItem(parameter)

    def _connect(self):
        self.refresh_button.clicked.connect(self.refresh_plot)



# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtWidgets import QMainWindow

# Own Imports
from kmap import __directory__

# Load .ui File
UI_file = __directory__ / 'ui/tabchoosewindow.ui'
TabChooseWindow_UI, _ = uic.loadUiType(UI_file)


class TabChooseWindow(QMainWindow, TabChooseWindow_UI):
    tabs_chosen = pyqtSignal(list)

    def __init__(self, sliced_tabs, orbital_tabs, *args, **kwargs):

        self.sliced_tabs = sliced_tabs
        self.orbital_tabs = orbital_tabs

        # Setup GUI
        super(TabChooseWindow, self).__init__(*args, **kwargs)
        self.setupUi(self)

        self.setAttribute(Qt.WA_DeleteOnClose, True)

        self._setup()
        self._connect()

        self.show()

    def choose_tabs(self):

        chosen_sliced_index = self.sliced_combobox.currentIndex()
        chosen_sliced_tab = self.sliced_tabs[chosen_sliced_index]

        chosen_orbital_index = self.orbital_combobox.currentIndex()
        chosen_orbital_tab = self.orbital_tabs[chosen_orbital_index]

        self.chosen_tabs = [chosen_sliced_tab, chosen_orbital_tab]

        self.close()

    def closeEvent(self, event):

        self.tabs_chosen.emit(self.chosen_tabs)
        self.deleteLater()
        event.accept()

    def _setup(self):

        for tab in self.sliced_tabs:
            self.sliced_combobox.addItem(tab.get_title())

        for tab in self.orbital_tabs:
            self.orbital_combobox.addItem(tab.get_title())

    def _connect(self):

        self.load.clicked.connect(self.choose_tabs)



# Third Party Imports
from lmfit import fit_report

# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtWidgets import QWidget
from PyQt5.QtCore import QDir, pyqtSignal

# Own Imports
from kmap import __directory__

# Load .ui File
UI_file = __directory__ / 'ui/lmfitresult.ui'
LMFitResult_UI, _ = uic.loadUiType(UI_file)


class LMFitResult(QWidget, LMFitResult_UI):
    print_triggered = pyqtSignal()
    cov_matrix_requested = pyqtSignal()
    plot_requested = pyqtSignal()

    def __init__(self, result, lmfit_model):

        self.result = result

        # Setup GUI
        super(LMFitResult, self).__init__()
        self.setupUi(self)
        self._setup(lmfit_model)
        self._connect()

    def get_fit_report(self):

        return fit_report(self.result)

    def get_covariance_matrix(self):

        return self.result.covar

    def _setup(self, lmfit_model):

        slices = lmfit_model.slice_policy[1]

        if len(list(slices)) == 1:
            type_ = 'Only One Slice (%i)' % slices[0]
            self.plot_button.setEnabled(False)

        else:
            if lmfit_model.slice_policy[2]:
                type_ = 'All Slices Combined'
                self.plot_button.setEnabled(False)

            else:
                type_ = 'All Slices Individually'

        self.type_label.setText(type_)

    def _connect(self):

        self.result_button.clicked.connect(self.print_triggered.emit)
        self.correlation_button.clicked.connect(self.cov_matrix_requested.emit)
        self.plot_button.clicked.connect(self.plot_requested.emit)



import numpy as np

from kmap.library.id import ID
from kmap.library.orbitaldata import OrbitalData
from kmap.config.config import config


class OrbitalDataTabModel():

    def __init__(self, controller):
        self.controller = controller

        self.displayed_plot_data = None
        self.orbitals = []

    def load_data_from_path(self, path):
        id_ = ID.new_ID()
        new_orbital = OrbitalData.init_from_file(path, ID=id_)
        self.orbitals.append([new_orbital, 'path', path, None, id_])

        return new_orbital

    def load_data_from_online(self, url, meta_data={}):
        id_ = ID.new_ID()
        new_orbital = OrbitalData.init_from_online(
            url, ID=id_, meta_data=meta_data)

        self.orbitals.append([new_orbital, 'url', url, meta_data, id_])

        return new_orbital

    def remove_data_by_object(self, orbital):
        index = [orb[0] for orb in self.orbitals].index(orbital)

        self.remove_data_by_index(index)

    def remove_data_by_index(self, index):
        del self.orbitals[index]

    def get_orbital_kmap_by_ID(self, ID):
        orbital = self.ID_to_orbital(ID)
        if orbital is None:
            raise IndexError('wrong ID')

        parameters = self.controller.get_parameters(ID)
        # Split of first element
        weight, *other = parameters
        s_share = float(config.get_key('orbital', 's_share'))
        # Get scaled kmap
        kmap = weight * orbital.get_kmap(*other, s_share=s_share)

        return kmap

    def update_displayed_plot_data(self):
        kmaps = []

        for orbital in self.orbitals:
            ID = orbital[0].ID

            if self.controller.get_use(ID):
                # Get all parameters for this orbital
                kmap = self.get_orbital_kmap_by_ID(ID)
                kmaps.append(kmap)

        if kmaps:
            # Sum kmaps
            self.displayed_plot_data = np.nansum(kmaps)

        else:
            self.displayed_plot_data = None

        return self.displayed_plot_data

    def remove_orbital_by_ID(self, ID):
        orbital = self.ID_to_orbital(ID)

        if orbital is not None:
            self.remove_data_by_object(orbital)

    def ID_to_orbital(self, ID):
        for orbital in self.orbitals:
            if orbital[0].ID == ID:
                return orbital[0]

        return None



import numpy as np
from pyqtgraph import PlotWidget, mkPen, mkBrush
from kmap.model.profileplot_model import ProfilePlotModel
from kmap.config.config import config
from kmap.library.plotdata import PlotData
from kmap.library.misc import normalize


class ProfilePlot(PlotWidget):

    def __init__(self, *args, **kwargs):
        super(ProfilePlot, self).__init__(*args, **kwargs)
        self._setup()

        self.plot_item = self.getPlotItem()
        self.model = ProfilePlotModel()
        self.title_suffixes = {'center': ' - Center Point',
                               'x': ' - Vertical Line',
                               'y': ' - Horizontal Line',
                               'roi': ' - Region of Interest',
                               'border': ' - Border of ROI',
                               'ring': ' - Annulus'}

    def clear(self):
        self.plot_item.clear()

    def get_data(self):
        data_sets = []

        for item in self.plot_item.listDataItems():
            name = item.name()
            x, y = item.getData()
            color = item.opts['pen'].color().getRgb()
            marker_size = item.opts['symbolSize']
            marker = item.opts['symbol']
            linewidth = item.opts['pen'].width()
            data_sets.append({'name': name, 'x': x, 'y': y, 'color': color,
                              'marker': marker, 'line width': linewidth,
                              'marker size': marker_size})

        return data_sets

    def plot(self, data, title, crosshair, region, phi_sample=720,
             line_sample=500, normalized=False):
        index = len(self.plot_item.listDataItems())
        colors = config.get_key('profile_plot', 'colors')
        color = colors.split(',')[index % len(colors)]
        line_width = int(config.get_key('profile_plot', 'line_width'))
        symbols = config.get_key('profile_plot', 'symbols')
        symbol = symbols.split(',')[index % len(symbols)]
        symbol_size = int(config.get_key('profile_plot', 'symbol_size'))

        if data is None:
            return

        elif isinstance(data, PlotData):
            x, y = self.model.get_plot_data(
                data, crosshair, region, phi_sample, line_sample)
        else:
            data, axis = data
            x = np.array(data.axes[axis].axis)
            y = []
            for i in range(len(x)):
                slice_ = data.slice_from_index(i, axis=axis)
                plot_data = crosshair.cut_from_data(slice_, region=region).data

                if np.isnan(plot_data).all():
                    y.append(np.nan)

                else:
                    if config.get_key('crosshair',
                                      'normalized_intensity') == 'True':
                        y.append(normalize(plot_data))

                    else:
                        y.append(np.nansum(plot_data))

            y = np.array(y)

        x = x[~np.isnan(y)]
        y = y[~np.isnan(y)]

        if normalized:
            y = y / max(y)

        self.plot_item.plot(x, y,
                            name=title + self.title_suffixes[region],
                            pen=mkPen(color, width=line_width),
                            symbol=symbol,
                            symbolPen=mkPen(color, width=symbol_size),
                            symbolBrush=mkBrush(color))

    def set_label(self, x, y):
        self.setLabel('left', text=y)
        self.setLabel('bottom', text=x)

    def _setup(self):
        self.addLegend()



# PyQt5 Imports
from PyQt5.QtWidgets import (
    QTreeWidgetItem, QCheckBox, QWidget, QLabel, QHBoxLayout, QLineEdit)
from PyQt5.QtCore import Qt, QObject, pyqtSignal

# Own Imports
from kmap.library.qwidgetsub import (
    CenteredLabel, WeightSpinBox,
    AngleSpinBox, EnergySpinBox, BackgroundSpinBox)


class SignalObject(QObject):

    value_changed = pyqtSignal()
    boundary_changed = pyqtSignal()
    expression_changed = pyqtSignal()
    vary_changed = pyqtSignal()


class LMFitTreeItem(QTreeWidgetItem):

    def __init__(self, tree):

        super().__init__(tree)

        self.children = []

    def is_vary(self):

        return self.vary.isChecked()

    def set_vary(self, state):

        self.vary.setChecked(state)

    def _setup(self, tree, name):

        self.name_label = CenteredLabel(name)

        self.vary = QCheckBox()
        layout = QHBoxLayout()
        layout.addWidget(self.vary)
        layout.setAlignment(Qt.AlignCenter)
        layout.setContentsMargins(0, 0, 0, 0)
        center_widget = QWidget()
        center_widget.setLayout(layout)

        tree.setItemWidget(self, 1, self.name_label)
        tree.setItemWidget(self, 3, center_widget)


class LMFitTreeTopLevelItem(LMFitTreeItem):

    def __init__(self, tree):

        super().__init__(tree)
        self.signals = SignalObject()

    def save_state(self):

        save = [child.save_state() for child in self.children]

        return save

    def restore_state(self, save):

        for i, child in enumerate(self.children):
            child.restore_state(save[i])

    def change_vary(self, state):

        for child in self.children:
            child.set_vary(state)

    def _change_to_matrix_state(self, state):

        self.vary.setEnabled(not state)
        
        for child in self.children:
            child._change_to_matrix_state(state)

    def _connect_child(self, child):
        # value_changed == redraw necessary
        child.signals.value_changed.connect(self.signals.value_changed.emit)
        child.signals.vary_changed.connect(self.signals.vary_changed.emit)

    def _connect(self):

        self.vary.stateChanged.connect(self.change_vary)

        for child in self.children:
            self._connect_child(child)


class BackgroundOptionsTreeItem(LMFitTreeTopLevelItem):

    def __init__(self, tree, parameters):

        super().__init__(tree)

        self._setup(tree, parameters)
        self._connect()

    def add_equation_parameter(self, tree, parameter):

        name = parameter.name
        if name not in self.parameters:
            item = BackgroundTreeItem(self, tree, parameter, 'Eq. Parameter')
            self._connect_child(item)
            self.children.append(item)
            self.parameters.append(name)

        else:
            index = self.parameters.index(name)
            item = self.children[index]
            item.update_parameter(parameter)

    def _change_to_matrix_state(self, state):

        super()._change_to_matrix_state(state)

    def _setup(self, tree, parameters):

        super()._setup(tree, 'Background Options')

        self.children = []
        self.parameters = []


class OrbitalOptionsTreeItem(LMFitTreeTopLevelItem):

    def __init__(self, tree, parameters):

        super().__init__(tree)

        self._setup(tree, parameters)
        self._connect()

    def _setup(self, tree, parameters):

        super()._setup(tree, 'Orbital Options')

        alpha = AngleTreeItem(self, tree, parameters['alpha'], 'Alpha')
        beta = AngleTreeItem(self, tree, parameters['beta'], 'Beta')
        energy = EnergyTreeItem(
            self, tree, parameters['E_kin'], 'Kinetic Energy')

        self.children = [alpha, beta, energy]
        self.parameters = ['alpha', 'beta', 'E_kin']


class OrbitalTreeItem(LMFitTreeTopLevelItem):

    def __init__(self, tree, orbital, parameters):

        super().__init__(tree)

        self.ID = orbital.ID
        self._setup(tree, orbital, parameters)
        self._connect()

    def _setup(self, tree, orbital, parameters):

        super()._setup(tree, orbital.name)

        self.ID_label = CenteredLabel('%i' % self.ID)
        self.name_label.setToolTip(str(orbital))
        tree.setItemWidget(self, 0, self.ID_label)

        weight = WeightTreeItem(
            self, tree, parameters['w_' + str(self.ID)], 'Weight')
        phi = AngleTreeItem(
            self, tree, parameters['phi_' + str(self.ID)], 'Phi')
        theta = AngleTreeItem(
            self, tree, parameters['theta_' + str(self.ID)], 'Theta')
        psi = AngleTreeItem(
            self, tree, parameters['psi_' + str(self.ID)], 'Psi')

        self.children = [weight, phi, theta, psi]


class LMFitDataTreeItem(LMFitTreeItem):

    def __init__(self, parent, parameter):

        self.parameter = parameter
        self.signals = SignalObject()

        super().__init__(parent)

    def save_state(self):

        save = {'vary': self.vary.isChecked(),
                'initial': self.initial_spinbox.value(),
                'min': self.min_spinbox.value(),
                'max': self.max_spinbox.value(),
                'expr': self.expression.text()}

        return save

    def restore_state(self, save):

        self.vary.setChecked(save['vary'])
        self.initial_spinbox.setValue(save['initial'])
        self.min_spinbox.setValue(save['min'])
        self.max_spinbox.setValue(save['max'])
        self.expression.setText(save['expr'])

    def _change_to_matrix_state(self, state):

        self.vary.setEnabled(not state)
        self.min_spinbox.setEnabled(not state)
        self.max_spinbox.setEnabled(not state)
        self.expression.setEnabled(not state)

    def _change_initial(self, value):

        self.parameter.value = value
        self.signals.value_changed.emit()

    def _change_min(self, value):

        self.parameter.min = value
        self.signals.boundary_changed.emit()

    def _change_max(self, value):

        self.parameter.max = value
        self.signals.boundary_changed.emit()

    def _change_expression(self, expression):

        self.parameter.expr = expression
        self.signals.expression_changed.emit()

    def _change_vary(self, state):

        self.parameter.vary = bool(state)
        self.signals.vary_changed.emit()

    def _setup(self, tree, name):

        name = self.parameter.name if name is None else name
        super()._setup(tree, name)

        self.alias_label = CenteredLabel(self.parameter.name)
        expression = '' if self.parameter.expr is None else self.parameter.expr
        self.expression = QLineEdit(expression)

        tree.setItemWidget(self, 2, self.alias_label)
        tree.setItemWidget(self, 4, self.initial_spinbox)
        tree.setItemWidget(self, 5, self.min_spinbox)
        tree.setItemWidget(self, 6, self.max_spinbox)
        tree.setItemWidget(self, 7, self.expression)

    def _connect(self):

        self.initial_spinbox.valueChanged.connect(self._change_initial)
        self.min_spinbox.valueChanged.connect(self._change_min)
        self.max_spinbox.valueChanged.connect(self._change_max)
        self.expression.returnPressed.connect(self._change_expression)
        self.vary.stateChanged.connect(self._change_vary)


class AngleTreeItem(LMFitDataTreeItem):

    def __init__(self, parent, tree, parameter, name=None):

        super().__init__(parent, parameter)

        self._setup(tree, name)
        self._connect()

    def _setup(self, tree, name):

        initial = self.parameter.value
        min_ = self.parameter.min
        max_ = self.parameter.max

        self.initial_spinbox = AngleSpinBox(initial, 'initial_spinbox')
        self.min_spinbox = AngleSpinBox(min_, 'min_spinbox')
        self.max_spinbox = AngleSpinBox(max_, 'max_spinbox')

        super()._setup(tree, name)


class BackgroundTreeItem(LMFitDataTreeItem):

    def __init__(self, parent, tree, parameter, name=None):

        super().__init__(parent, parameter)

        self._setup(tree, name)
        self._connect()

    def update_parameter(self, parameter):

        initial = self.parameter.value
        min_ = self.parameter.min
        max_ = self.parameter.max
        expr = self.parameter.expr
        vary = self.parameter.vary

        self.parameter = parameter

        self._change_initial(initial)
        self._change_min(min_)
        self._change_max(max_)
        self._change_expression(expr)
        self._change_vary(vary)

    def _change_to_matrix_state(self, state):

        super()._change_to_matrix_state(state)

        if state and self.parameter.name == 'c':
            if self.initial_spinbox.value() == 0.:
                self.initial_spinbox.setValue(1.0)

            self.vary.setEnabled(True)

    def _setup(self, tree, name):

        initial = self.parameter.value
        min_ = self.parameter.min
        max_ = self.parameter.max

        self.initial_spinbox = BackgroundSpinBox(value=initial)
        self.min_spinbox = BackgroundSpinBox(value=min_)
        self.max_spinbox = BackgroundSpinBox(value=max_)

        super()._setup(tree, name)


class EnergyTreeItem(LMFitDataTreeItem):

    def __init__(self, parent, tree, parameter, name=None):

        super().__init__(parent, parameter)

        self._setup(tree, name)
        self._connect()

    def _setup(self, tree, name):

        initial = self.parameter.value
        min_ = self.parameter.min
        max_ = self.parameter.max

        self.initial_spinbox = EnergySpinBox(value=initial)
        self.min_spinbox = EnergySpinBox(value=min_)
        self.max_spinbox = EnergySpinBox(value=max_)

        super()._setup(tree, name)


class WeightTreeItem(LMFitDataTreeItem):

    def __init__(self, parent, tree, parameter, name=None):

        super().__init__(parent, parameter)

        self._setup(tree, name)
        self._connect()

    def _change_to_matrix_state(self, state):

        super()._change_to_matrix_state(state)
        self.vary.setEnabled(True)

    def _setup(self, tree, name):

        initial = self.parameter.value
        min_ = self.parameter.min
        max_ = self.parameter.max

        self.initial_spinbox = WeightSpinBox(value=initial)
        self.min_spinbox = WeightSpinBox(value=min_)
        self.max_spinbox = WeightSpinBox(value=max_)

        super()._setup(tree, name)


class LMFitResultTreeItem(QTreeWidgetItem):

    def __init__(self, parent):

        super().__init__(parent)

        self.children = []
        self.parameters = []

    def update_result(self, result):

        for child, param in zip(self.children, self.parameters):
            child.update_result(result[param])

    def _setup(self, tree, name):

        self.name_label = CenteredLabel(name)

        tree.setItemWidget(self, 1, self.name_label)


class BackgroundOptionsResultTreeItem(LMFitResultTreeItem):

    def __init__(self, tree, result, background_variables=[]):

        super().__init__(tree)

        self._setup(tree, result, background_variables)

    def _setup(self, tree, result, background_variables):

        super()._setup(tree, 'Background Options')

        self.parameters = []
        self.children = []

        for key in background_variables:
            item = DataResultTreeItem(self, tree, result[key],
                                      'Eq. Parameter', units='',
                                      decimals=1)
            self.parameters.append(key)
            self.children.append(item)


class OrbitalOptionsResultTreeItem(LMFitResultTreeItem):

    def __init__(self, tree, result):

        super().__init__(tree)

        self._setup(tree, result)

    def _setup(self, tree, result):

        super()._setup(tree, 'Orbital Options')

        alpha = DataResultTreeItem(self, tree, result['alpha'],
                                   'Alpha', units='°', decimals=3)
        beta = DataResultTreeItem(self, tree, result['beta'],
                                  'Beta', units='°', decimals=3)
        energy = DataResultTreeItem(self, tree, result['E_kin'],
                                    'Kinetic Engery', units='  eV',
                                    decimals=2)

        self.parameters = ['alpha', 'beta', 'E_kin']
        self.children = [alpha, beta, energy]


class OrbitalResultTreeItem(LMFitResultTreeItem):

    def __init__(self, tree, orbital, result):

        super().__init__(tree)

        self._setup(tree, orbital, result)

    def _setup(self, tree, orbital, result):

        super()._setup(tree, orbital.name)
        self.name_label.setToolTip(str(orbital))

        self.ID = orbital.ID
        self.ID_label = CenteredLabel('%i' % self.ID)
        tree.setItemWidget(self, 0, self.ID_label)

        weight = DataResultTreeItem(self, tree,
                                    result['w_' + str(self.ID)], 'Weight',
                                    units='', decimals=2)
        phi = DataResultTreeItem(self, tree,
                                 result['phi_' + str(self.ID)], 'Phi',
                                 units='°', decimals=3)
        theta = DataResultTreeItem(self, tree,
                                   result['theta_' + str(self.ID)], 'Theta',
                                   units='°', decimals=3)
        psi = DataResultTreeItem(self, tree,
                                 result['psi_' + str(self.ID)], 'Psi',
                                 units='°', decimals=3)

        self.parameters = [
            variable + '_' + str(self.ID) for variable in
            ['w', 'phi', 'theta', 'psi']]
        self.children = [weight, phi, theta, psi]


class DataResultTreeItem(LMFitResultTreeItem):

    def __init__(self, parent, tree, parameter, name, units, decimals):

        self.units = units
        self.decimals = decimals

        super().__init__(parent)
        self._setup(tree, parameter, name)

    def update_result(self, parameter):

        self.parameter = parameter
        self._update_text()

    def _update_text(self):

        value = self.parameter.value
        stderr = self.parameter.stderr
        alias = self.parameter.name

        value = '%.{0}f%s'.format(self.decimals) % (value, self.units)

        if stderr is None:
            stderr = '-'

        else:
            stderr = '%.{0}f%s'.format(self.decimals) % (stderr, self.units)

        self.alias_label.setText(alias)
        self.result.setText(value)
        self.stderr.setText(stderr)

    def _setup(self, tree, parameter, name):

        super()._setup(tree, name)

        self.alias_label = CenteredLabel()
        self.result = CenteredLabel('')
        self.stderr = CenteredLabel('')

        tree.setItemWidget(self, 2, self.alias_label)
        tree.setItemWidget(self, 3, self.result)
        tree.setItemWidget(self, 4, self.stderr)

        self.update_result(parameter)



# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtWidgets import QWidget

# Third Party Imports
import numpy as np
from pyqtgraph import ImageView, PlotItem, AxisItem

# Own Imports
from kmap.model.pyqtgraphplot_model import PyQtGraphPlotModel
from kmap.library.axis import Axis
from kmap.config.config import config


class PyQtGraphPlot(ImageView):

    def __init__(self, *args, plot_data=None, **kwargs):

        # Setup GUI
        self.plot_view = PlotItem()
        super(PyQtGraphPlot, self).__init__(
            *args, view=self.plot_view, **kwargs)
        self._setup()

        self.model = PyQtGraphPlotModel(plot_data)

        self.refresh_plot()

    def plot(self, plot_data):

        self.model.plot_data = plot_data

        self.refresh_plot()

    def refresh_plot(self, keep_max_level=False):

        self.clear()

        if self.model.plot_data is None:
            return

        image, pos, scale, range_ = self.model.get_plot()

        if np.all(np.isnan(image)) == True:
            return

        old_level = self.get_levels()

        # Plot
        self.setImage(image, autoRange=True,
                      autoLevels=True, pos=pos, scale=scale)

        self.set_range(range_)
        ratio = (range_[1][1] - range_[1][0]) / (range_[0][1] - range_[0][0])

        if config.get_key('pyqtgraph', 'fixed_ratio') == 'True':
            self.set_aspect_ratio(ratio)

        else:
            self.set_aspect_ratio(None)

        if (keep_max_level or
                config.get_key('pyqtgraph', 'keep_max_level') == 'True'):
            levels = [np.nanmin(image.data), np.nanmax(image.data)]

            # Catch empty plots
            if old_level != (0, 1):
                levels[0] = old_level[0] if old_level[0] < levels[0] else \
                levels[0]
                levels[1] = old_level[1] if old_level[1] > levels[1] else \
                levels[1]

            self.set_levels(levels)

    def set_range(self, range_):

        padding = float(config.get_key('pyqtgraph', 'padding'))
        self.view.setRange(xRange=range_[0], yRange=range_[1],
                           update=True, padding=padding)

    def set_levels(self, levels):

        self.getHistogramWidget().setLevels(*levels)
        self.setHistogramRange(*levels)

    def get_levels(self):

        return self.getHistogramWidget().getLevels()

    def set_aspect_ratio(self, ratio):

        if ratio is not None:
            self.plot_view.setAspectLocked(True, ratio=ratio)

        else:
            self.plot_view.setAspectLocked(False)

    def set_labels(self, x, y):

        color = config.get_key('pyqtgraph', 'axis_color')
        size = config.get_key('pyqtgraph', 'axis_size')

        if isinstance(x, list):
            self.set_label('bottom', x[0], x[1], color, size)

        elif isinstance(x, Axis):
            self.set_label('bottom', x.label, x.units, color, size)

        else:
            self.set_label('bottom', str(x), None, color, size)

        if isinstance(y, list):
            self.set_label('left', y[0], y[1], color, size)

        elif isinstance(y, Axis):
            self.set_label('left', y.label, y.units, color, size)

        else:
            self.set_label('left', str(y), None, color, size)

    def set_label(self, side, label, units=None, color='k', size=1):

        axis = AxisItem(side, text=label, units=units,
                        **{'color': color, 'font-size': size})

        if config.get_key('pyqtgraph', 'show_axis_label') == 'True':
            axis.showLabel(True)

        else:
            axis.showLabel(False)

        self.view.setAxisItems({side: axis})

    def get_plot_data(self):

        return self.model.plot_data

    def get_LUT(self):

        colormap = self.getHistogramWidget().gradient.colorMap()
        nPts = int(config.get_key('pyqtgraph', 'nPts'))
        LUT = colormap.getLookupTable(mode='float', alpha=True, nPts=nPts)

        return LUT

    def get_label(self, side):

        return self.plot_view.getAxis(side).label.toHtml()

    def _setup(self):

        self.view.invertY(False)
        self.view.hideButtons()
        self.ui.roiBtn.hide()
        self.ui.menuBtn.hide()



# Python Imports
from math import ceil, floor

# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QMainWindow, QHBoxLayout, QWidget
from matplotlib.backends.backend_qt5agg import (FigureCanvas,
                                                NavigationToolbar2QT)
from matplotlib.colors import ListedColormap
from matplotlib.figure import Figure
# Third party Imports
from matplotlib.ticker import AutoMinorLocator

# Own Imports
from kmap import __directory__
from kmap.library.qwidgetsub import AspectWidget
from kmap.config.config import config
from kmap.model.matplotlibwindow_model import (MatplotlibImageModel,
                                               MatplotlibLineModel)

# Load .ui File
UI_file = __directory__ / 'ui/matplotlibwindow.ui'
MatplotlibWindow_UI, _ = uic.loadUiType(UI_file)


class MatplotlibWindow(QMainWindow):

    def __init__(self):
        super(MatplotlibWindow, self).__init__()

    def update_canvas(self):
        self.figure.canvas.draw()

    def update_x_range(self, range_):
        self.axes.set_xlim(range_)

        self.update_canvas()

    def update_y_range(self, range_):
        self.axes.set_ylim(range_)

        self.update_canvas()

    def add_grid(self, which='No Grid'):
        if which == 'No Grid':
            self.axes.grid(False, which='both')

        elif which == 'Major Only':
            self.axes.grid(True, which='major')
            self.axes.grid(False, which='minor')

        elif which == 'Major and Minor':
            self.axes.grid(True, which='both')

        self.update_canvas()

    def add_minor_ticks(self, num):
        # Is off by one, no idea why
        self.axes.xaxis.set_minor_locator(AutoMinorLocator(num + 1))
        self.axes.yaxis.set_minor_locator(AutoMinorLocator(num + 1))

        self.update_canvas()

    def add_x_label(self, label):
        self.axes.set_xlabel(label)

        self.update_canvas()

    def add_y_label(self, label):
        self.axes.set_ylabel(label)

        self.update_canvas()

    def add_title(self, title):
        self.axes.set_title(title)

        self.update_canvas()

    def display(self):
        pass

    def closeEvent(self, event):
        # Catch closing to close options window as well
        try:
            self.options.close()
            del self.options

        except:
            pass

        finally:
            event.accept()


class MatplotlibImageWindow(MatplotlibWindow, MatplotlibWindow_UI):

    def __init__(self, plot_data, LUT=None):
        self.model = MatplotlibImageModel(plot_data)
        self.LUT = ListedColormap(LUT, 'cm_user')

        super(MatplotlibImageWindow, self).__init__()
        self.setupUi(self)

        self._setup()
        self._connect()

        self.display()

        self.update_canvas()
        self.fit_axis()

        self.show()
        self.options.show()

    def display(self):
        x, y, image = self.model.x, self.model.y, self.model.image

        self.plot = self.axes.pcolormesh(x, y, image, cmap=self.LUT)

    def fit_axis(self):
        # New Limits. Round to second decimal place to always fit entire
        # image
        x, y = self.model.x, self.model.y
        x_limit = [floor(min(x) * 100) / 100,
                   ceil(max(x) * 100) / 100]
        y_limit = [floor(min(y) * 100) / 100,
                   ceil(max(y) * 100) / 100]

        self.options.set_x_range(x_limit)
        self.options.set_y_range(y_limit)

        self.update_x_range(x_limit)
        self.update_y_range(y_limit)

        self.update_canvas()

    def add_colorbar(self, enable):
        if enable:
            self.colorbar = self.figure.colorbar(self.plot, ax=self.axes)

        else:
            self.colorbar.remove()

        self.update_canvas()

    def _setup(self):
        self.figure = Figure()
        self.axes = self.figure.add_subplot(111)

        # Canvas
        canvas = FigureCanvas(self.figure)
        self.addToolBar(NavigationToolbar2QT(canvas, self))

        # Main Layout
        layout = QHBoxLayout()
        layout.addWidget(canvas)

        aux = config.get_key('matplotlib', 'ratio')
        ratio = 0 if aux == 'None' else float(aux)
        self.central_widget = AspectWidget(ratio=ratio)
        self.central_widget.setLayout(layout)

        self.setCentralWidget(self.central_widget)

        self.options = MatplotlibImageOptions()

    def _connect(self):
        self.options.colorbar_changed.connect(self.add_colorbar)
        self.options.grid_changed.connect(self.add_grid)
        self.options.ticks_changed.connect(self.add_minor_ticks)
        self.options.title_changed.connect(self.add_title)
        self.options.x_label_changed.connect(self.add_x_label)
        self.options.y_label_changed.connect(self.add_y_label)
        self.options.x_range_changed.connect(self.update_x_range)
        self.options.y_range_changed.connect(self.update_y_range)
        self.options.fit_axis_triggered.connect(self.fit_axis)


class MatplotlibLineWindow(MatplotlibWindow, MatplotlibWindow_UI):

    def __init__(self, plot_data):
        self.model = MatplotlibLineModel(plot_data)

        super(MatplotlibLineWindow, self).__init__()
        self.setupUi(self)

        self._setup()
        self._connect()

        self.display()

        self.update_canvas()

        self.show()
        self.options.show()

    def display(self):
        data = self.model.data

        for data_set in data:
            name = data_set['name']
            x = data_set['x']
            y = data_set['y']
            color = [c / 255 for c in data_set['color']]
            marker = data_set['marker']
            marker = '*' if marker == 'star' else marker
            line_width = data_set['line width']
            marker_size = data_set['marker size']
            self.plot = self.axes.plot(x, y, color=color, marker=marker,
                                       linewidth=line_width,
                                       markersize=marker_size, label=name)

        self.axes.legend()

    def _setup(self):
        self.figure = Figure()
        self.axes = self.figure.add_subplot(111)

        # Canvas
        canvas = FigureCanvas(self.figure)
        self.addToolBar(NavigationToolbar2QT(canvas, self))

        # Main Layout
        layout = QHBoxLayout()
        layout.addWidget(canvas)

        self.central_widget = AspectWidget()
        self.central_widget.setLayout(layout)

        self.setCentralWidget(self.central_widget)

        self.options = MatplotlibLineOptions()

    def _connect(self):
        self.options.grid_changed.connect(self.add_grid)
        self.options.ticks_changed.connect(self.add_minor_ticks)
        self.options.title_changed.connect(self.add_title)
        self.options.x_label_changed.connect(self.add_x_label)
        self.options.y_label_changed.connect(self.add_y_label)
        self.options.x_range_changed.connect(self.update_x_range)
        self.options.y_range_changed.connect(self.update_y_range)


class MatplotlibOptions(QWidget):
    grid_changed = pyqtSignal(str)
    ticks_changed = pyqtSignal(int)
    title_changed = pyqtSignal(str)
    x_label_changed = pyqtSignal(str)
    y_label_changed = pyqtSignal(str)
    x_range_changed = pyqtSignal(list)
    y_range_changed = pyqtSignal(list)

    def __init__(self):
        # Setup GUI
        super(MatplotlibOptions, self).__init__()

    def get_x_range(self):
        min_ = self.x_min_spinbox.value()
        max_ = self.x_max_spinbox.value()

        return [min_, max_]

    def get_y_range(self):
        min_ = self.y_min_spinbox.value()
        max_ = self.y_max_spinbox.value()

        return [min_, max_]

    def set_x_range(self, range_):
        min_, max_ = range_

        self.x_min_spinbox.blockSignals(True)
        self.x_max_spinbox.blockSignals(True)

        self.x_min_spinbox.setMaximum(min_ + 1)
        self.x_min_spinbox.setValue(min_)

        self.x_max_spinbox.setMinimum(max_ - 1)
        self.x_max_spinbox.setValue(max_)

        self.x_min_spinbox.blockSignals(False)
        self.x_max_spinbox.blockSignals(False)

        self._update_boundaries()

    def set_y_range(self, range_):
        min_, max_ = range_

        self.y_min_spinbox.blockSignals(True)
        self.y_max_spinbox.blockSignals(True)

        self.y_min_spinbox.setMaximum(min_ + 1)
        self.y_min_spinbox.setValue(min_)

        self.y_max_spinbox.setMinimum(max_ - 1)
        self.y_max_spinbox.setValue(max_)

        self.y_min_spinbox.blockSignals(False)
        self.y_max_spinbox.blockSignals(False)

        self._update_boundaries()

    def add_grid(self, which):
        self.grid_changed.emit(which)

    def add_ticks(self, num):
        self.ticks_changed.emit(num)

    def add_title(self, text):
        self.title_changed.emit(text)

    def add_x_label(self, text):
        self.x_label_changed.emit(text)

    def add_y_label(self, text):
        self.y_label_changed.emit(text)

    def change_x_range(self):
        range_ = self.get_x_range()

        self._update_boundaries()

        self.x_range_changed.emit(range_)

    def change_y_range(self):
        range_ = self.get_y_range()

        self._update_boundaries()

        self.y_range_changed.emit(range_)

    def _update_boundaries(self):
        self.x_min_spinbox.setMaximum(self.x_max_spinbox.value() - 0.01)
        self.x_max_spinbox.setMinimum(self.x_min_spinbox.value() + 0.01)
        self.y_min_spinbox.setMaximum(self.y_max_spinbox.value() - 0.01)
        self.y_max_spinbox.setMinimum(self.y_min_spinbox.value() + 0.01)

    def _connect(self):
        self.grid_combobox.currentTextChanged.connect(self.add_grid)
        self.ticks_spinbox.valueChanged.connect(self.add_ticks)
        self.title_line_edit.textChanged.connect(self.add_title)
        self.x_line_edit.textChanged.connect(self.add_x_label)
        self.y_line_edit.textChanged.connect(self.add_y_label)
        self.x_min_spinbox.valueChanged.connect(self.change_x_range)
        self.x_max_spinbox.valueChanged.connect(self.change_x_range)
        self.y_min_spinbox.valueChanged.connect(self.change_y_range)
        self.y_max_spinbox.valueChanged.connect(self.change_y_range)


# Load .ui File
UI_file = __directory__ / 'ui/matplotlibimageoptions.ui'
MatplotlibImageOptions_UI, _ = uic.loadUiType(UI_file)


class MatplotlibImageOptions(MatplotlibOptions, MatplotlibImageOptions_UI):
    colorbar_changed = pyqtSignal(int)
    fit_axis_triggered = pyqtSignal()

    def __init__(self):
        # Setup GUI
        super(MatplotlibImageOptions, self).__init__()
        self.setupUi(self)
        self._connect()

    def add_colorbar(self, state):
        enable = True if state != 0 else False
        self.colorbar_changed.emit(enable)

    def fit_axis(self):
        self.fit_axis_triggered.emit()

    def _connect(self):
        super()._connect()
        self.colorbar_checkbox.stateChanged.connect(self.add_colorbar)
        self.fit_button.clicked.connect(self.fit_axis)


# Load .ui File
UI_file = __directory__ / 'ui/matplotliblineoptions.ui'
MatplotlibLineOptions_UI, _ = uic.loadUiType(UI_file)


class MatplotlibLineOptions(MatplotlibOptions, MatplotlibLineOptions_UI):

    def __init__(self):
        # Setup GUI
        super(MatplotlibLineOptions, self).__init__()
        self.setupUi(self)
        self._connect()



# Python Imports
import numpy as np

# Own Imports
from kmap.config.config import config


class PyQtGraphPlotModel():

    def __init__(self, plot_data):

        self.plot_data = plot_data

    def get_plot(self):

        image = self.plot_data.data
        scale = self.plot_data.step_size
        pixel_center = config.get_key('pyqtgraph', 'pixel_center') == 'True'
        pos = self._calculate_pos(scale, pixel_center=pixel_center)
        range_ = self._calculate_range(scale, pixel_center=pixel_center)

        return image, pos, scale, range_

    def _calculate_pos(self, scale, pixel_center=True):

        if pixel_center:
            pos = self.plot_data.range[:, 0] - scale / 2
        else:
            pos = self.plot_data.range[:, 0]

        return pos

    def _calculate_range(self, scale, pixel_center=True):

        old_range = self.plot_data.range

        if pixel_center:
            shift = np.array([[-1, 1], [-1, 1]]) * scale / 2
            new_range = old_range + shift

        else:
            new_range = old_range

        return new_range



import os
import logging
import logging.config
from pathlib import Path

from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import QApplication

import pyqtgraph as pg
import matplotlib as plt

from kmap.config.config import config
from kmap.controller.mainwindow import MainWindow
from kmap import __version__, __project__, __directory__


class kMap(QApplication):

    def __init__(self, sysarg):
        # Apply various configurations
        self.load_settings(startup=True)

        logging.getLogger('kmap').debug(
            'Initializing kMap.py v' + __version__ + '.')

        # Initialize application
        super().__init__(sysarg)
        self.setApplicationVersion(__version__)
        self.setApplicationName(__project__)
        self.setDesktopFileName(__project__)

        if config.get_key('app', 'dark_mode') == 'True':
            try:
                import qdarkstyle
                self.setStyleSheet(qdarkstyle.load_stylesheet(qt_api='pyqt5'))
            except ImportError:
                print(
                    'Dark Theme could not be activated. Please reinstall kMap.py or install "qdarkstyle" manually.')

    def run(self):
        logging.getLogger('kmap').info('Starting up kMap.')

        # Creating mainwindow
        self.main_window = MainWindow()

        super().exec_()

    def test(self):
        pass

    def load_settings(self, startup=False):
        # Logging
        # Delete old log files if user set to do so
        if startup and config.get_key('logging', 'persistent') == 'False':
            log_file = __directory__ / '../default.log'
            if os.path.exists(log_file):
                os.remove(log_file)

        logging.config.fileConfig(config.get_config(
            'logging'), disable_existing_loggers=False)

        # PyQtGraph
        value = config.get_key('pyqtgraph', 'background')
        if value == 'None':
            pg.setConfigOption('background', None)

        else:
            pg.setConfigOption('background', value)

        value = config.get_key('pyqtgraph', 'foreground')
        if value == 'None':
            pg.setConfigOption('foreground', None)

        else:
            pg.setConfigOption('foreground', value)

        value = config.get_key('pyqtgraph', 'antialias')
        if value == 'True':
            pg.setConfigOption('antialias', True)

        else:
            pg.setConfigOption('antialias', False)

        value = config.get_key('pyqtgraph', 'imageAxisOrder')
        pg.setConfigOption('imageAxisOrder', value)

        self.setFont(QFont(config.get_key('font', 'font'), int(
            config.get_key('font', 'size')), QFont.Normal))

        logging.getLogger('kmap').debug('Settings loaded successfully.')

        # MatPlotlib
        path = config.get_key('paths', 'matplotlib')
        if path != 'None':
            path = __directory__ / path
            plt.rcParams['savefig.directory'] = str(path)



"""Defines the PlotData class.

This file defines a class named PlotData designed to hold arbitrary
image-like. For this purpose it holds a data matrix and axes data as
well as some other useful properties. It also comes with interpolations 
and Gaussian filter methods.
"""

# Python Imports
from numbers import Number

# Third Party Imports
import numpy as np
import itertools as it
from scipy.ndimage import gaussian_filter, rotate
from scipy.interpolate import RegularGridInterpolator as RGI

# Own Imports
from kmap.library.misc import axis_from_range, range_from_axes


class PlotData():
    """
    Basic data class holding necessary information for plotting.

    Defines a data class for data designed to be plotted, including
    interpolation and smoothing.
    """

    def __init__(self, data, range_):
        """
        Args:
            data (float): 2D array of numbers with at least 2 elements
                in each dimension. Can contain NaN values.
            range_ (float): List of two lists containing the min and max
                value of x_axis and y_axis respectively. Can not contain
                NaN values and min value as to be strictly smaller than
                max value. (e.g. [[x_min, x_max], [y_min, y_max]])

        Public Attributes:
            data (float): See args.
            range (float): See args.
            x_axis (float): 1D array for the first axis of the data.
            y_axis (float): 1D array for the second axis of the data.
            step_size (float) : List of step sizes for each axes.
        """

        # Set data
        self.data = np.array(data, dtype=np.float64)

        if (self.data.ndim != 2 or len(self.data) < 2 or
                len(self.data[0]) < 2):
            raise TypeError('data has to be 2D array with at least 2 ' +
                            'elements in each dimension')

        else:
            self.data[~ np.isfinite(self.data)] = np.nan

        # Set range
        self.range = np.array(range_, dtype=np.float64)

        if self.range.shape != (2, 2):
            raise TypeError('range has to be of shape (2, 2)')

        if not np.isfinite(self.range).all():
            raise ValueError('range can only have finite values')

        # Set axes
        self.x_axis = axis_from_range(self.range[0], self.data.shape[1])
        self.y_axis = axis_from_range(self.range[1], self.data.shape[0])

        # Set step_size
        _, self.step_size = range_from_axes(self.x_axis, self.y_axis)

    def interpolate(self, x_axis, y_axis, update=False, **kwargs):
        """Wrapper method to interpolate data at entire grid specified
            by the axes passed.

        Args:
            x_axis (float): Same es as constructor.
            y_axis (float): Same es as constructor.
            update (bool): If True data and axis saved in this instance
                will be overriden with interpolation result. If False
                interpolate will only return the result but data will
                stay unchanged.
            **kwargs: See interpolate_points.

        Returns:
            (PlotData): PlotData object with the interpolated data.
        """

        points = np.array(list(it.product(y_axis, x_axis)))
        new_data = self.interpolate_points(
            points[:, 1], points[:, 0], **kwargs)
        new_data = np.array(new_data, dtype=np.float64).reshape(
            len(y_axis), len(x_axis))

        if update:
            self.data, self.x_axis, self.y_axis = new_data, x_axis, y_axis
            self.range, self.step_size = range_from_axes(x_axis, y_axis)

            return self

        else:
            range_, _ = range_from_axes(x_axis, y_axis)
            new_plot_data = PlotData(new_data, range_)

            return new_plot_data

    def interpolate_points(self, x, y, interpolator='rgi',
                           bounds_error=False, **kwargs):
        """Interpolates the current data at the points
            [(x_1,y_1), (x_2,y_2),...].

        Args:
            x_axis (float): Same es as constructor.
            y_axis (float): Same es as constructor.
            interpolator (string): Specifies the underlying
                interpolation method to be used. Currently available:
                    rgi - RegularGridInterpolator from SciPy
            bounds_error (bool): If True interpolating outside current
                range will result in an error. If False data points
                outside will be filled with np.nan.
            **kwargs: Will be passed to the interpolation
                method, thus pass any arguments the chosen interpolation
                accepts.

        Returns:
            (float): List of data interpolated at the specified points.

        """

        if interpolator == 'rgi':
            points = np.transpose([y, x])
            rgi = RGI((self.y_axis, self.x_axis), self.data,
                      bounds_error=bounds_error, **kwargs)

            intensities = np.array(rgi(points), dtype=np.float64)

        else:
            NotImplementedError('Chosen interpolator is unknown')

        return intensities

    def symmetrise(self, symmetry='no', mirror=False, update=False):
        """Symmetrises the data set. Optional mirroring available.

        Args:
            symmetry (str): '2-fold', '3-fold', '4-fold' or 'no'.
            mirror (bool): Whether the data gets mirrored as well. Along first
                axis for 2- and 3-fold and along first and second axis for 4-
                fold symmetry.
            update (bool): If True data in this instance
                will be overriden with symmetrised result. If False
                this function will only return the result but data will
                stay unchanged.

        Returns:
            (PlotData): PlotData object with the symmetrised data.

        """

        new_data = self.data

        if symmetry == '2-fold':
            new_data += rotate(np.nan_to_num(new_data), 180, reshape=False)

            if mirror:
                new_data += np.flip(new_data, 0)
                new_data /= 2

            new_data /= 2


        elif symmetry == '3-fold':
            aux_1 = rotate(np.nan_to_num(new_data), 120, reshape=False)
            aux_2 = rotate(np.nan_to_num(new_data), 240, reshape=False)
            new_data += aux_1 + aux_2

            if mirror:
                new_data += np.flip(new_data, 0)
                new_data /= 2

            new_data /= 3

        elif symmetry == '4-fold':
            aux_1 = rotate(np.nan_to_num(new_data), 90, reshape=False)
            aux_2 = rotate(np.nan_to_num(new_data), 180, reshape=False)
            aux_3 = rotate(np.nan_to_num(new_data), 270, reshape=False)
            new_data += aux_1 + aux_2 + aux_3

            if mirror:
                new_data += np.flip(new_data, 0)
                new_data += np.flip(new_data, 1)
                new_data /= 2

            new_data /= 4

        if update:
            self.data = new_data

            return self

        else:
            new_plot_data = PlotData(new_data, self.range)

            return new_plot_data

    def smooth(self, sigma_x, sigma_y, mode='nearest',
               update=False, fill_value=np.nan, **kwargs):
        """Applies a 2D Gaussian filter to the current data. Wrapper
            method for the "gaussian_filter" method from them 
            "scipy.ndimage" module.

        Args:
            sigma_x (float): See scipy.ndimage.gaussian_filter
                documentation.
            sigma_y (float): See scipy.ndimage.gaussian_filter
                documentation.
            mode (string): See scipy.ndimage.gaussian_filter
                documentation. Changed default to "nearest".
            update (bool): If True data and axis saved in this instance
                will be overriden with smoothing result. If False
                smoothing will only return the result but data will
                stay unchanged.
            fill_value (float): All NaN values in the data will be
                temporarily filled with value specified in "fill_value"
                for the smoothing procedure. (This can be necessary
                since NaN values will "infect" other values around it
                during smoothing)
            **kwargs: See scipy.ndimage.gaussian_filter

        Returns:
            (PlotData): PlotData object with the smoothed data.

        """

        # Gaussian filter needs number of pixel
        sigma_x = round(sigma_x / self.step_size[0])
        sigma_y = round(sigma_y / self.step_size[1])

        fill_mask = np.zeros(self.data.shape, dtype=bool)
        fill_mask[np.isnan(self.data)] = True
        self.data[fill_mask] = fill_value

        new_data = gaussian_filter(
            self.data, [sigma_y, sigma_x], mode=mode, **kwargs)
        new_data[fill_mask] = np.nan

        if update:
            self.data = new_data

            return self

        else:
            new_plot_data = PlotData(new_data, self.range)

            return new_plot_data

    def copy(self):
        data = self.data.copy()
        range_ = self.range.copy()
        return PlotData(data, range_)

    def __add__(self, other):
        if isinstance(other, self.__class__):
            if ((self.x_axis == other.x_axis).all() and
                    (self.y_axis == other.y_axis).all()):
                return PlotData(self.data + other.data, self.range)

            else:
                raise ValueError('Axes need to be equal')

        elif isinstance(other, np.ndarray):
            return PlotData(self.data + other, self.range)

        elif isinstance(other, float) or isinstance(other, int):
            return PlotData(self.data + other, self.range)

        else:
            raise TypeError('Can not add \'%s\' to PlotData' % other.__class__)

    def __sub__(self, other):
        if isinstance(other, self.__class__):
            if ((self.x_axis == other.x_axis).all() and
                    (self.y_axis == other.y_axis).all()):
                return PlotData(self.data - other.data, self.range)

            else:
                raise ValueError('Axes need to be equal')

        elif isinstance(other, np.ndarray):
            return PlotData(self.data - other, self.range)

        elif isinstance(other, float) or isinstance(other, int):
            return PlotData(self.data - other, self.range)

        else:
            raise TypeError('Can not subtract \'%s\' to PlotData' %
                            other.__class__)

    def __mul__(self, other):
        if isinstance(other, self.__class__):
            if ((self.x_axis == other.x_axis).all() and
                    (self.y_axis == other.y_axis).all()):
                return PlotData(self.data * other.data, self.range)

            else:
                raise ValueError('Axes need to be equal')

        elif isinstance(other, float) or isinstance(other, int):
            return PlotData(self.data * other, self.range)

        else:
            raise TypeError('Can not multiply \'%s\' to PlotData' %
                            other.__class__)

    def __rmul__(self, other):
        if isinstance(other, self.__class__):
            if ((self.x_axis == other.x_axis).all() and
                    (self.y_axis == other.y_axis).all()):
                return PlotData(self.data * other.data, self.range)

            else:
                raise ValueError('Axes need to be equal')

        elif isinstance(other, float) or isinstance(other, int):
            return PlotData(self.data * other, self.range)

        else:
            raise TypeError('Can not multiply \'%s\' to PlotData' %
                            other.__class__)

    def __pow__(self, other):
        if isinstance(other, float) or isinstance(other, int):
            return PlotData(self.data * self.data, self.range)

        else:
            raise TypeError(
                'Can take PlotData to the power of \'%s\'' % other.__class__)



# Third Party Imports
import numpy as np

# Own Imports
from kmap.library.misc import axis_from_range


class Axis():

    def __init__(self, label, units, range_, num):

        self.label = label
        self.units = units
        self.range = range_
        self.axis = axis_from_range(range_, num)
        self.stepsize = abs(self.axis[1] - self.axis[0])
        self.num = len(self.axis)

    @classmethod
    def init_from_hdf_list(cls, axis, num):

        if Axis._is_correct_axis(axis):
            return cls(*axis, num)

    @classmethod
    def _is_correct_axis(self, axis):

        if not isinstance(axis, list) or len(axis) != 3:
            raise ValueError('axis has to be a list of length 3')

        label, units, range_ = axis
        if not isinstance(label, str) or not label:
            raise ValueError('axis label has to be a non empty string')

        if not isinstance(units, str) or not units:
            raise ValueError('axis units has to be a non empty string')

        if len(range_) != 2:
            raise ValueError('axis range has to be list of length 2')

        minimum, maximum = range_
        if not np.isfinite(minimum) or not np.isfinite(maximum):
            raise ValueError('axis range can not contain inf or nan values')

        if not minimum < maximum:
            raise ValueError(
                'axis minimum has to strictly smaller than maxmimum')

        return True

    def __str__(self):

        rep = 'Label:\t%s\nUnits:\t%s\nRange:\t[%.3f, %.3f]' % (
            self.label, self.units, *self.range)
        return rep



# Python Imports
import logging

# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtWidgets import QWidget, QHeaderView, QHBoxLayout, QSizePolicy

# Own Imports
from kmap import __directory__
from kmap.controller.orbitaltablerow import OrbitalTableRow
from kmap.controller.lmfittreeitems import (
    OrbitalTreeItem, OrbitalOptionsTreeItem, LMFitDataTreeItem,
    OrbitalOptionsResultTreeItem, OrbitalResultTreeItem,
    DataResultTreeItem, BackgroundOptionsTreeItem,
    BackgroundOptionsResultTreeItem)

# Load .ui File
UI_file = __directory__ / 'ui/lmfittree.ui'
LMFitTree_UI, _ = uic.loadUiType(UI_file)


class LMFitBaseTree(QWidget):
    item_selected = pyqtSignal()

    def get_selected_orbital_ID(self):

        selected_items = self.tree.selectedItems()

        for item in selected_items:
            if (isinstance(item, OrbitalTreeItem) or
                    isinstance(item, OrbitalResultTreeItem)):
                return item.ID

            elif ((isinstance(item, LMFitDataTreeItem) or
                   isinstance(item, DataResultTreeItem)) and
                  (isinstance(item.parent(), OrbitalTreeItem) or
                   isinstance(item.parent(), OrbitalResultTreeItem))):
                return item.parent().ID

        return -1

    def _connect(self):

        self.tree.itemSelectionChanged.connect(self.item_selected.emit)


class LMFitTree(LMFitBaseTree, LMFitTree_UI):
    value_changed = pyqtSignal()
    vary_changed = pyqtSignal()

    def __init__(self, orbitals, parameters, *args, **kwargs):

        # Setup GUI
        super(LMFitTree, self).__init__(*args, **kwargs)
        self.setupUi(self)
        self._setup(orbitals, parameters)
        self._connect()

    def save_state(self):

        save = [self.tree.topLevelItem(i).save_state()
                for i in range(self.tree.topLevelItemCount())]

        return save

    def restore_state(self, save):

        for i in range(self.tree.topLevelItemCount()):
            self.tree.topLevelItem(i).restore_state(save[i])

    def add_equation_parameter(self, parameter):

        self.background_item.add_equation_parameter(self.tree, parameter)

    def _change_to_matrix_state(self, state):

        for i in range(self.tree.topLevelItemCount()):
            self.tree.topLevelItem(i)._change_to_matrix_state(state)

    def _setup(self, orbitals, parameters):

        widths = [60, 0, 100, 80, 130, 130, 130, 200]

        for col, width in enumerate(widths):
            self.tree.setColumnWidth(col, width)

        self.tree.header().setResizeMode(1, QHeaderView.Stretch)
        self.tree.header().setDefaultAlignment(Qt.AlignCenter)

        # Add TreeItems
        self.orbital_options_item = OrbitalOptionsTreeItem(
            self.tree, parameters)
        self.background_item = BackgroundOptionsTreeItem(self.tree, parameters)
        self.tree.addTopLevelItem(self.orbital_options_item)
        for orbital in orbitals:
            self.tree.addTopLevelItem(
                OrbitalTreeItem(self.tree, orbital, parameters))

    def _connect(self):

        LMFitBaseTree._connect(self)

        for i in range(self.tree.topLevelItemCount()):
            item = self.tree.topLevelItem(i)
            item.signals.value_changed.connect(self.value_changed.emit)
            item.signals.vary_changed.connect(self.vary_changed.emit)


# Load .ui File
UI_file = __directory__ / 'ui/lmfitresulttree.ui'
LMFitResultTree_UI, _ = uic.loadUiType(UI_file)


class LMFitResultTree(LMFitBaseTree, LMFitResultTree_UI):

    def __init__(self, orbitals, result, background_variables=[]):

        # Setup GUI
        super(LMFitResultTree, self).__init__()
        self.setupUi(self)
        self._setup(orbitals, result, background_variables)
        self._connect()

    def update_result(self, result):

        for i in range(self.tree.topLevelItemCount()):
            item = self.tree.topLevelItem(i)
            item.update_result(result)

    def _setup(self, orbitals, result, background_variables):

        widths = [60, 0, 100, 150, 150]

        for col, width in enumerate(widths):
            self.tree.setColumnWidth(col, width)

        self.tree.header().setResizeMode(1, QHeaderView.Stretch)
        self.tree.header().setDefaultAlignment(Qt.AlignCenter)

        # Add TreeItems
        self.tree.addTopLevelItem(OrbitalOptionsResultTreeItem(
            self.tree, result))
        self.tree.addTopLevelItem(BackgroundOptionsResultTreeItem(
            self.tree, result, background_variables))
        for orbital in orbitals:
            self.tree.addTopLevelItem(
                OrbitalResultTreeItem(self.tree, orbital, result))



"""Defines the LMFit tab and the LMFitResult tab.

This file defines two similar types of tabs: the LMFit and the
LMFitResult tab as well as a common base class LMFitBaseTab.
"""
# Python Imports
import logging

# Third Party Imports
import h5py
import numpy as np

# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtCore import pyqtSignal, Qt
from PyQt5.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QFileDialog

# Own Imports
from kmap import __directory__
from kmap.library.axis import Axis
from kmap.library.qwidgetsub import Tab
from kmap.model.lmfit_model import LMFitModel
from kmap.controller.dataslider import DataSlider
from kmap.controller.colormap import Colormap
from kmap.controller.crosshairannulus import CrosshairAnnulus
from kmap.controller.interpolation import LMFitInterpolation
from kmap.controller.lmfittree import LMFitTree, LMFitResultTree
from kmap.controller.matplotlibwindow import MatplotlibImageWindow
from kmap.controller.lmfitresult import LMFitResult
from kmap.controller.lmfitoptions import LMFitOptions
from kmap.controller.lmfitorbitaloptions import LMFitOrbitalOptions
from kmap.config.config import config

# Load .ui File
UI_file = __directory__ / 'ui/lmfittab.ui'
LMFitTab_UI, _ = uic.loadUiType(UI_file)


class LMFitBaseTab(Tab):

    def refresh_sliced_plot(self):
        index = self.slider.get_index()
        kmap = self.model.get_sliced_kmap(index)

        self.sliced_plot.plot(kmap)

    def refresh_selected_plot(self, param=None):
        ID = self.tree.get_selected_orbital_ID()

        if ID == -1:
            self.selected_plot.plot(None)

        else:
            kmap = self.model.get_orbital_kmap(ID, param)
            self.selected_plot.plot(kmap)

    def refresh_sum_plot(self, param=None):
        kmap = self.model.get_weighted_sum_kmap(param)
        self.sum_plot.plot(kmap)

        return kmap

    def refresh_residual_plot(self, param=None, weight_sum_data=None):
        index = self.slider.get_index()
        residual = self.model.get_residual(index, param, weight_sum_data)

        level = np.nanmax(np.absolute(residual.data))

        if config.get_key('pyqtgraph', 'keep_max_level_residual') == 'True':
            old_level = np.nanmax(np.absolute(self.residual_plot.get_levels()))

            if old_level != 1 and old_level > level:
                level = old_level

        self.residual_plot.plot(residual)

        self.residual_plot.set_levels([-level, level])

        self.update_chi2_label(weight_sum_data)

    def refresh_all(self):

        self.refresh_sliced_plot()
        self.refresh_selected_plot()
        kmap = self.refresh_sum_plot()
        self.refresh_residual_plot(weight_sum_data=kmap)

    def update_chi2_label(self, weight_sum_data=None):
        slice_index = self.slider.get_index()
        reduced_chi2 = self.model.get_reduced_chi2(
            slice_index, weight_sum_data)
        self.residual_label.setText('Residual (red. Chi^2: %.3E)'
                                    % reduced_chi2)

    def display_in_matplotlib(self):
        windows = []

        for plot in [self.residual_plot, self.sum_plot]:
            data = plot.model.plot_data
            LUT = plot.get_LUT()
            windows.append(MatplotlibImageWindow(data, LUT=LUT))

        return windows

    def transpose(self, constant_axis):
        self.model.transpose(constant_axis)
        self.refresh_sliced_plot()
        self.refresh_residual_plot()

    def change_symmetry(self, symmetry, mirror):
        self.model.set_sliced_symmetrization(symmetry, mirror)
        self.refresh_sliced_plot()
        self.refresh_residual_plot()

    def closeEvent(self, event):
        del self.model

        Tab.closeEvent(self, event)

    def _refresh_orbital_plots(self):
        self.refresh_selected_plot()
        kmap = self.refresh_sum_plot()
        self.refresh_residual_plot(weight_sum_data=kmap)

    def _setup(self):
        self.slider = DataSlider(self.model.sliced_data)
        self.crosshair = CrosshairAnnulus(self.residual_plot)
        self.colormap = Colormap(
            [self.sliced_plot, self.selected_plot, self.sum_plot])
        residual_colormap = Colormap(self.residual_plot)

        residual_colormap.set_colormap('blueAndRed')

    def _connect(self):
        self.crosshair.crosshair_changed.connect(self.crosshair.update_label)
        self.crosshair.crosshair_changed.connect(self.refresh_all)

        self.slider.slice_changed.connect(self.change_slice)
        self.slider.axis_changed.connect(self.change_slice)
        self.slider.tranpose_triggered.connect(self.transpose)
        self.slider.symmetry_changed.connect(self.change_symmetry)

        self.tree.item_selected.connect(self.refresh_selected_plot)


class LMFitTab(LMFitBaseTab, LMFitTab_UI):
    fit_finished = pyqtSignal(list, dict)

    def __init__(self, sliced_tab, orbital_tab):
        self.sliced_tab = sliced_tab
        self.orbital_tab = orbital_tab
        self.model = LMFitModel(sliced_tab.get_data(),
                                orbital_tab.get_orbitals())
        s_share = float(config.get_key('orbital', 's_share'))
        self.model.set_s_share(s_share)

        # Setup GUI
        super(LMFitTab, self).__init__()
        self.setupUi(self)

        self._setup()
        self._connect()

        self.refresh_all()

    @classmethod
    def init_from_save(cls, save, sliced_data, orbitals):
        tab = LMFitTab(sliced_data, orbitals)
        tab.title = save['title']

        tab.slider.restore_state(save['slider']),
        tab.crosshair.restore_state(save['crosshair']),
        tab.orbital_options.restore_state(save['orbital_options']),
        tab.interpolation.restore_state(save['interpolation']),
        tab.lmfit_options.restore_state(save['lmfit_options'])
        tab.tree.restore_state(save['tree'])

        return tab

    def get_title(self):
        return self.title

    def get_data(self):
        return [self.model.sliced_data, self.model.orbitals]

    def trigger_fit(self):
        try:
            results = self.model.fit()

        except ValueError as e:
            logging.getLogger('kmap').warning(str(e))
            self.lmfit_options.update_fit_button()

            return

        settings = self.model.get_settings()

        self.fit_finished.emit(results, settings)

    def change_slice(self):
        axis_index = self.slider.get_axis()
        slice_policy = self.lmfit_options.get_slice_policy()
        combined = True if slice_policy == 'all combined' else False
        slice_indices = (self.slider.get_index()
                         if slice_policy == 'only one' else 'all')

        self.model.set_slices(
            slice_indices, axis_index=axis_index, combined=combined)

        self.refresh_sliced_plot()
        self.refresh_residual_plot()

    def change_axis(self):
        axis = self.interpolation.get_axis()
        self.model.set_axis(axis)

        self.refresh_all()

    def save_state(self):
        save = {'title': self.title,
                'slider': self.slider.save_state(),
                'crosshair': self.crosshair.save_state(),
                'orbital_options': self.orbital_options.save_state(),
                'interpolation': self.interpolation.save_state(),
                'lmfit_options': self.lmfit_options.save_state(),
                'tree': self.tree.save_state()}

        return save, [self.sliced_tab, self.orbital_tab]

    def _change_slice_policy(self, slice_policy):
        axis = self.slider.get_axis()

        if slice_policy == 'all':
            self.model.set_slices('all', axis_index=axis, combined=False)

        elif slice_policy == 'only one':
            index = self.slider.get_index()
            self.model.set_slices([index], axis_index=axis, combined=False)

        else:
            self.model.set_slices('all', axis_index=axis, combined=True)

    def _change_method(self, method):

        self._change_to_matrix_state(method == 'matrix_inversion')

        self.model.set_fit_method(method)

    def _change_to_matrix_state(self, state):

        if state:
            variables = self.model.background_equation[1]
            if 'c' not in variables:
                self.lmfit_options._pre_factor_background()

        self.tree._change_to_matrix_state(state)

    def _change_region(self, *args):
        self.model.set_region(*args)

        self.refresh_all()

    def _change_background(self, *args):

        new_variables = self.model.set_background_equation(*args)
        for variable in new_variables:
            self.tree.add_equation_parameter(variable)

        self.refresh_sum_plot()
        self.refresh_residual_plot()

    def _setup(self):
        LMFitBaseTab._setup(self)

        self.orbital_options = LMFitOrbitalOptions()
        self.tree = LMFitTree(self.model.orbitals, self.model.parameters)
        self.interpolation = LMFitInterpolation()
        self.lmfit_options = LMFitOptions(self)

        self.change_axis()
        self.model.set_crosshair(self.crosshair.model)
        self._change_background(self.lmfit_options.get_background())

        layout = QVBoxLayout()
        layout.setContentsMargins(3, 3, 3, 3)
        layout.setSpacing(3)
        self.scroll_area.widget().setLayout(layout)
        layout.insertWidget(0, self.slider)
        layout.insertWidget(1, self.orbital_options)
        layout.insertWidget(2, self.interpolation)
        layout.insertWidget(3, self.lmfit_options)
        layout.insertWidget(4, self.colormap)
        layout.insertWidget(5, self.crosshair)

        self.layout.insertWidget(1, self.tree)

    def _connect(self):
        LMFitBaseTab._connect(self)

        self.interpolation.interpolation_changed.connect(self.change_axis)

        self.tree.value_changed.connect(self._refresh_orbital_plots)
        self.tree.vary_changed.connect(self.update_chi2_label)
        self.lmfit_options.background_changed.connect(self._change_background)
        self.lmfit_options.fit_triggered.connect(self.trigger_fit)
        self.lmfit_options.method_changed.connect(self._change_method)
        self.lmfit_options.slice_policy_changed.connect(
            self._change_slice_policy)
        self.lmfit_options.region_changed.connect(self._change_region)

        self.orbital_options.symmetrization_changed.connect(
            self.model.set_symmetrization)
        self.orbital_options.symmetrization_changed.connect(
            self._refresh_orbital_plots)
        self.orbital_options.polarization_changed.connect(
            self.model.set_polarization)
        self.orbital_options.polarization_changed.connect(
            self._refresh_orbital_plots)


class LMFitResultTab(LMFitBaseTab, LMFitTab_UI):
    open_plot_tab = pyqtSignal(list, list, Axis, list, list)

    def __init__(self, lmfit_tab, results, settings):
        self.results = results
        self.lmfit_tab = lmfit_tab
        self.current_result = self.results[0]
        self.settings = settings

        self.model = LMFitModel(*self.lmfit_tab.get_data())
        self.model.set_settings(settings)

        # Setup GUI
        super(LMFitResultTab, self).__init__()
        self.setupUi(self)

        self._setup()
        self._connect()

        self.refresh_all()

    @classmethod
    def init_from_save(cls, save, tab, ID_map=None):
        results = save['results']

        if ID_map is not None:
            for result in results:
                result = LMFitResultTab._apply_ID_map(result[1], ID_map)

        settings = save['settings']

        tab = LMFitResultTab(tab, results, settings)

        return tab

    @classmethod
    def _apply_ID_map(cls, result, ID_map):
        params = result.params
        new_params = params.copy()

        for map_ in ID_map:
            for parameter in params.items():
                old_name, parameter = parameter

                if old_name.endswith(str(map_[0])):
                    new_name = '_'.join(old_name.split(
                        '_')[:-1]) + '_' + str(map_[1])
                    new_params[new_name] = new_params[old_name]

                    if old_name != new_name:
                        del new_params[old_name]

        result.params = new_params
        return result

    def save_state(self):
        save = {'title': self.title,
                'results': self.results,
                'settings': self.settings}

        return save, [self.lmfit_tab]

    def get_title(self):
        return 'Results'

    def change_slice(self):
        index = self.slider.get_index()

        self.current_result = self.results[0] if len(
            self.results) == 1 else self.results[index]

        self.update_tree()
        self.result.result = self.current_result[1]
        self.refresh_sliced_plot()
        self._refresh_orbital_plots()

    def refresh_selected_plot(self):
        params = self.current_result[1].params
        super().refresh_selected_plot(params)

    def refresh_sum_plot(self):
        params = self.current_result[1].params
        return super().refresh_sum_plot(params)

    def refresh_residual_plot(self, weight_sum_data=None):
        params = self.current_result[1].params
        super().refresh_residual_plot(params, weight_sum_data)

    def update_tree(self):
        params = self.current_result[1].params
        self.tree.update_result(params)

    def print_result(self):
        report = self.result.get_fit_report()

        print(report)

    def print_covariance_matrix(self):
        cov_matrix = self.result.get_covariance_matrix()

        print(cov_matrix)

    def get_orbitals(self):
        return self.model.orbitals

    def plot(self):
        results = [result[1] for result in self.results]
        orbitals = self.model.orbitals
        axis = self.model.sliced_data.axes[self.model.slice_policy[0]]
        kmaps = abs(self.get_residual_kmaps())
        residuals = list(np.nansum(np.nansum(kmaps, axis=1), axis=1))

        self.open_plot_tab.emit(results, orbitals, axis, residuals,
                                self.model.background_equation[1])

    def export_to_hdf5(self):
        path = config.get_key('paths', 'hdf5_export_start')
        if path == 'None':
            file_name, _ = QFileDialog.getSaveFileName(
                None, 'Save .hdf5 File (*.hdf5)')
        else:
            start_path = str(__directory__ / path)
            file_name, _ = QFileDialog.getSaveFileName(
                None, 'Save .hdf5 File (*.hdf5)', str(start_path))

        if not file_name:
            return
        else:
            h5file = h5py.File(file_name, 'w')

        kmaps = self.get_residual_kmaps()

        slice_axis = self.slider.data.axes[0]
        x_axis = self.slider.data.axes[1]
        y_axis = self.slider.data.axes[2]

        h5file.create_dataset('name', data='Residual')
        h5file.create_dataset('axis_1_label', data=slice_axis.label)
        h5file.create_dataset('axis_2_label', data=x_axis.label)
        h5file.create_dataset('axis_3_label', data=y_axis.label)
        h5file.create_dataset('axis_1_units', data=slice_axis.units)
        h5file.create_dataset('axis_2_units', data=x_axis.units)
        h5file.create_dataset('axis_3_units', data=y_axis.units)
        h5file.create_dataset('axis_1_range', data=slice_axis.range)
        h5file.create_dataset('axis_2_range', data=x_axis.range)
        h5file.create_dataset('axis_3_range', data=y_axis.range)
        h5file.create_dataset('data', data=kmaps, dtype='f8',
                              compression='gzip', compression_opts=9)
        h5file.close()

    def get_residual_kmaps(self):
        kmaps = []
        for i, result in enumerate(self.results):
            residual = self.model.get_residual(i, result[1].params)
            kmaps.append(residual.data)
        kmaps = np.array(kmaps)

        return kmaps

    def _setup(self):
        LMFitBaseTab._setup(self)

        self.result = LMFitResult(self.current_result[1], self.model)
        self.tree = LMFitResultTree(
            self.model.orbitals, self.current_result[1].params,
            self.model.background_equation[1])
        self.crosshair._set_model(self.model.crosshair)

        layout = QVBoxLayout()
        self.scroll_area.widget().setLayout(layout)
        layout.insertWidget(0, self.slider)
        layout.insertWidget(1, self.result)
        layout.insertWidget(3, self.colormap)
        layout.insertWidget(4, self.crosshair)

        self.layout.insertWidget(1, self.tree)

    def _connect(self):
        self.result.print_triggered.connect(self.print_result)
        self.result.cov_matrix_requested.connect(self.print_covariance_matrix)
        self.result.plot_requested.connect(self.plot)

        LMFitBaseTab._connect(self)



# PyQt5 Imports
from PyQt5 import uic
from PyQt5.QtWidgets import QWidget
from PyQt5.QtCore import Qt

# Own Imports
from kmap import __directory__

# Load .ui File
UI_file = __directory__ / 'ui/sliceddatabaseoptions.ui'
SlicedDataBaseOptions_UI, _ = uic.loadUiType(UI_file)


class SlicedDataBaseOptions(QWidget, SlicedDataBaseOptions_UI):

    def __init__(self, *args, **kwargs):

        # Setup GUI
        super(SlicedDataBaseOptions, self).__init__(*args, **kwargs)
        self.setupUi(self)
        self._setup()

    def get_parameters(self):

        name = self.line_edit.text()
        name = name if name else 'no name given'
        polarization = self._get_polarization()
        symmetry = self._get_symmetry()
        orientation = self._get_orientation()
        other = self._get_other()

        return (name, *other, *orientation, *polarization, symmetry)

    def _get_orientation(self):

        phi = self.phi_spinbox.value()
        theta = self.theta_spinbox.value()
        psi = self.psi_spinbox.value()

        return phi, theta, psi

    def _get_other(self):

        photon = self.photon_spinbox.value()
        fermi = self.fermi_spinbox.value()
        broadening = self.broadening_spinbox.value()
        dk = self.dk_spinbox.value()

        return photon, fermi, broadening, dk

    def _get_polarization(self):

        Ak_index = self.polarization_combobox.currentIndex()
        polarization = 'p'
        if Ak_index == 0:
            Ak_type = 'no'

        elif Ak_index == 1:
            Ak_type = 'toroid'

        else:
            Ak_type = 'NanoESCA'

            if Ak_index == 3:
                polarization = 's'

            elif Ak_index == 4:
                polarization = 'unpolarized'

            elif Ak_index == 5:
                polarization = 'C+'

            elif Ak_index == 6:
                polarization = 'C-'

            elif Ak_index == 7:
                polarization = 'CDAD'

        alpha = self.alpha_spinbox.value()
        beta = self.beta_spinbox.value()
        gamma = 'auto'

        return Ak_type, polarization, alpha, beta, gamma

    def closeEvent(self, event):

        self.deleteLater()
        event.accept()

    def _get_symmetry(self):

        index = self.symmetrization_combobox.currentIndex()
        available_symmetries = ['no', '2-fold', '2-fold+mirror',
                                '3-fold', '3-fold+mirror', '4-fold',
                                '4-fold+mirror']

        return available_symmetries[index]

    def _setup(self):

        self.setWindowFlags(Qt.CustomizeWindowHint | Qt.WindowTitleHint)



import numpy as np
from pyqtgraph import PlotWidget, mkPen, mkBrush
from kmap.config.config import config


class LMFitPlot(PlotWidget):

    def __init__(self, *args, **kwargs):
        super(LMFitPlot, self).__init__(*args, **kwargs)
        self._setup()

        self.plot_item = self.getPlotItem()

    def clear(self):
        self.plot_item.clear()

    def get_data(self):
        data_sets = []

        for item in self.plot_item.listDataItems():
            name = item.name()
            x, y = item.getData()
            color = item.opts['pen'].color().getRgb()
            marker_size = item.opts['symbolSize']
            marker = item.opts['symbol']
            linewidth = item.opts['pen'].width()
            data_sets.append({'name': name, 'x': x, 'y': y, 'color': color,
                              'marker': marker, 'line width': linewidth,
                              'marker size': marker_size})

        return data_sets

    def plot(self, x, y, title):
        index = len(self.plot_item.listDataItems())
        colors = config.get_key('profile_plot', 'colors')
        colors = colors.split(',')
        color = colors[index % len(colors)]
        line_width = int(config.get_key('profile_plot', 'line_width'))
        symbols = config.get_key('profile_plot', 'symbols')
        symbols = symbols.split(',')
        symbol = symbols[index % len(symbols)]
        symbol_size = int(config.get_key('profile_plot', 'symbol_size'))

        self.plot_item.plot(x, y,
                            name=title,
                            pen=mkPen(color, width=line_width),
                            symbol=symbol,
                            symbolPen=mkPen(color, width=symbol_size),
                            symbolBrush=mkBrush(color))

    def set_label(self, x, y):
        self.setLabel('left', text=y)
        self.setLabel('bottom', text=x)

    def _setup(self):
        pass
        self.addLegend()
